from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtWidgets import QColorDialog, QFrame, QHBoxLayout, QLabel, QWidget

from angrmanagement.data.object_container import ObjectContainer

if TYPE_CHECKING:
    from PySide6.QtGui import QColor


class QColorOption(QWidget):
    def __init__(self, color: QColor, label: str, parent=None) -> None:
        super().__init__(parent=parent)

        self.color = ObjectContainer(color, "The current color")
        self.label = label

        self._init_widgets()

    def set_color(self, color) -> None:
        self.color.am_obj = color
        self.color.am_event()

    def mouseReleaseEvent(self, event) -> None:  # pylint:disable=unused-argument
        dialog = QColorDialog()
        dialog.setCurrentColor(self.color.am_obj)
        dialog.exec()
        if dialog.result() == QColorDialog.Accepted:
            self.set_color(dialog.currentColor())

    def _init_widgets(self) -> None:
        layout = QHBoxLayout()
        frame = QFrame()
        frame.setFixedWidth(30)
        frame.setFixedHeight(15)

        def update_color(**kwargs) -> None:  # pylint:disable=unused-argument
            r, g, b, a = self.color.getRgb()
            frame.setStyleSheet(f"background-color: rgba({r},{g},{b},{a});")

        update_color()
        self.color.am_subscribe(update_color)

        text = QLabel(self.label)

        layout.addWidget(frame)
        layout.addWidget(text)
        self.setLayout(layout)



from __future__ import annotations

from pygments.lexers.c_cpp import CLexer
from pyqodeng.core.api import CodeEdit
from pyqodeng.core.modes import PygmentsSyntaxHighlighter
from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QPushButton,
)

from angrmanagement.config import Conf
from angrmanagement.logic import GlobalInfo
from angrmanagement.ui.dialogs.xref import XRefDialog
from angrmanagement.ui.widgets.qccode_edit import ColorSchemeIDA
from angrmanagement.utils.layout import add_to_grid


class FunctionDialog(QDialog):
    """
    Dialog displaying information about a Function.
    """

    def __init__(self, function, parent=None) -> None:
        super().__init__(parent)
        self.workspace = GlobalInfo.main_window.workspace
        self.function = function
        self._init_widgets()
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.setWindowTitle(f"Function {self.function.name}")
        self.setMinimumWidth(600)
        self.adjustSize()

    def _init_widgets(self) -> None:
        font = QFont(Conf.disasm_font)

        main_layout = QGridLayout()
        self.setLayout(main_layout)

        r = 0

        for label, text in [
            ("Name:", self.function.name),
            ("Address:", f"{self.function.addr:x}"),
            ("Binary:", f"{self.function.binary}"),
            ("Offset:", f"{self.function.offset:x}"),
            (
                "Calling Convention:",
                "<Unknown>" if self.function.calling_convention is None else f"{self.function.calling_convention}",
            ),
            ("Tags:", ", ".join(self.function.tags)),
            ("Cyclomatic Complexity:", str(self.function.cyclomatic_complexity)),
        ]:
            main_layout.addWidget(QLabel(label), r, 0)
            le = QLineEdit(text, self)
            le.setFont(font)
            le.setReadOnly(True)  # FIXME: Support editing
            main_layout.addWidget(le, r, 1)
            r += 1

        main_layout.addWidget(QLabel("C Prototype:"), r, 0)
        ce = CodeEdit(self)
        if self.function.prototype is not None:
            decl = self.function.prototype.c_repr(full=True).replace("()", self.function.name, 1) + ";"
            ce.document().setPlainText(decl)
        ce.modes.append(PygmentsSyntaxHighlighter(ce.document(), CLexer(), ColorSchemeIDA()))
        ce.setFixedHeight(ce.fontMetrics().height() * 2)
        ce.setReadOnly(True)  # FIXME: Support editing
        main_layout.addWidget(ce, r, 1)
        r += 1

        attrs_group_box = QGroupBox("Attributes")
        attrs_layout = QGridLayout()
        attrs_group_box.setLayout(attrs_layout)
        main_layout.addWidget(attrs_group_box, r, 0, 1, 2)
        attrs = []
        for label, checked in [
            ("Alignment", self.function.is_alignment),
            ("PLT", self.function.is_plt),
            ("SimProcedure", self.function.is_simprocedure),
            ("Returning", self.function.returning),
            ("Prototype Guessed", self.function.is_prototype_guessed),
            ("Variadic", self.function.prototype is not None and self.function.prototype.variadic),
        ]:
            cb = QCheckBox(label, self)
            cb.setChecked(checked)
            cb.setEnabled(False)  # FIXME: Support editing
            attrs.append(cb)
        add_to_grid(attrs_layout, 3, attrs)
        r += 1

        actions_group_box = QGroupBox("Actions")
        actions_layout = QGridLayout()
        actions_group_box.setLayout(actions_layout)
        main_layout.addWidget(actions_group_box, r, 0, 1, 2)
        actions = [
            ("&Decompile", self._decompile),
            ("Dis&assemble", self._disassemble),
            ("Show &Xrefs", self._show_xrefs),
            ("Show &Strings", self._show_strings),
            # FIXME: Create Call State
            # FIXME: Show Function Documentation
            # FIXME: Show In Proximity View
            # FIXME: Plugin extras
        ]
        action_buttons = []
        for label, handler in actions:
            btn = QPushButton(label, self)
            btn.clicked.connect(handler)
            action_buttons.append(btn)
        add_to_grid(actions_layout, 4, action_buttons)
        r += 1

        main_layout.setRowStretch(r, 1)
        r += 1

    def _decompile(self) -> None:
        self.workspace.decompile_function(self.function)
        self.accept()

    def _disassemble(self) -> None:
        self.workspace.jump_to(self.function.addr)
        self.accept()

    def _show_xrefs(self) -> None:
        # FIXME: Make this an action on workspace
        dialog = XRefDialog(
            addr=self.function.addr,
            dst_addr=self.function.addr,
            xrefs_manager=self.workspace.main_instance.project.kb.xrefs,
            instance=self.workspace.main_instance,
            parent=self,
        )
        dialog.exec_()
        self.accept()

    def _show_strings(self) -> None:
        self.workspace.show_strings_view()
        view = self.workspace.view_manager.first_view_in_category("strings")
        if view is not None:
            view.select_function(self.function)
        self.accept()



from __future__ import annotations

from angrmanagement.ui.widgets.qfeature_map import QFeatureMap

from .toolbar import Toolbar
from .toolbar_dock import ToolBarDockWidget


class FeatureMapToolbar(Toolbar):
    """
    Displays the current feature map. Monitors most recently focused view via instance.activate_view_state to show
    cursors on the feature map, and allows selection in the feature map to control location of active view.
    """

    def __init__(self, window) -> None:
        super().__init__(window, "Feature Map")
        self._toolbar = None
        self._feature_map = None
        self._is_subscribed = False

    def _subscribe_events(self) -> None:
        self.window.workspace.main_instance.active_view_state.am_subscribe(self._on_view_state_updated)
        self._feature_map.addr.am_subscribe(self._on_feature_map_addr_selected)
        self._is_subscribed = True

    def _unsubscribe_events(self) -> None:
        if self._is_subscribed:
            self.window.workspace.main_instance.active_view_state.am_unsubscribe(self._on_view_state_updated)
            self._feature_map.addr.am_unsubscribe(self._on_feature_map_addr_selected)

    def qtoolbar(self):
        if self._toolbar is None:
            self._feature_map = QFeatureMap(self.window.workspace.main_instance, parent=self.window)
            self._toolbar = ToolBarDockWidget(self._feature_map, "Feature Map", parent=None)
            self._subscribe_events()
        return self._toolbar

    def shutdown(self) -> None:
        self._unsubscribe_events()

    def _on_feature_map_addr_selected(self) -> None:
        target_view = self.window.workspace.view_manager.most_recently_focused_view
        if hasattr(target_view, "jump_to"):
            addr = self._feature_map.addr.am_obj
            if addr is not None:
                target_view.jump_to(addr)

    def _on_view_state_updated(self) -> None:
        vs = self.window.workspace.main_instance.active_view_state
        if vs.am_none:
            return
        self._feature_map.set_cursor_addrs(vs.cursors)



from __future__ import annotations

from PySide6.QtCore import QObject, Signal

from angrmanagement.data.object_container import EventSentinel


def _make_config_entry_value(init_value, type_):
    class ConfigurationEntryValue(QObject, EventSentinel):
        """
        An object container that both calls am_event() and emits 'changed' on an update
        This object allows usage of both types of event triggers
        """

        changed = Signal(type_)

        def __init__(self, value) -> None:
            QObject.__init__(self)
            if not hasattr(self, "am_subscribers"):
                EventSentinel.__init__(self)
            self._value = value

        def get(self):
            return self._value

        def set(self, value) -> None:
            """
            Set the value, if it changes, trigger events
            """
            if value != self._value:
                self._value = value
                self.am_event()
                self.changed.emit(value)

    return ConfigurationEntryValue(init_value)


class ConfigurationEntry:
    """
    Describes a configuration entry in angr management.
    """

    __slots__ = ("name", "type_", "_value", "default_value")

    def __init__(self, name: str, type_, value, default_value=None) -> None:
        self.name = name
        self.type_ = type_
        self._value = _make_config_entry_value(value, self.type_)
        self.default_value = default_value

    def copy(self):
        """
        Copies over data, does *not* copy subscribers, signal connections, or slot connections
        """
        return ConfigurationEntry(self.name, self.type_, self.value, default_value=self.default_value)

    @property
    def value(self):
        return self._value.get()

    @value.setter
    def value(self, new) -> None:
        self._value.set(new)

    #
    # Event functions
    #

    @property
    def changed(self) -> Signal:
        return self._value.changed

    def subscribe(self, listener) -> None:
        self._value.am_subscribe(listener)

    def unsubscribe(self, listener) -> None:
        self._value.am_unsubscribe(listener)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QEvent, QPoint, QPointF, QRectF, QSize, Qt
from PySide6.QtGui import QBrush, QColor, QLinearGradient, QPen, QPolygonF
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsLineItem,
    QGraphicsPolygonItem,
    QGraphicsRectItem,
    QGraphicsScene,
    QGraphicsView,
    QHBoxLayout,
    QWidget,
)

from angrmanagement.config import Conf
from angrmanagement.logic.debugger import DebuggerWatcher
from angrmanagement.logic.debugger.bintrace import BintraceDebugger

if TYPE_CHECKING:
    from collections.abc import Sequence

    from angrmanagement.data.instance import Instance


class TraceMapItem(QGraphicsItem):
    """
    Trace map item to be rendered in graphics scene.
    """

    ZVALUE_CHECKPOINT = 1
    ZVALUE_ADDR = 2
    ZVALUE_HOVER = 3

    def __init__(self, instance: Instance) -> None:
        super().__init__()
        self.instance = instance

        self._width: int = 1
        self._height: int = 1

        self._addr: int = 0
        self._indicator_items: Sequence[QGraphicsItem] = []

        self._hover_addr: int | None = None
        self._hover_items: Sequence[QGraphicsItem] = []

        self._checkpoints = []
        self._checkpoint_items: Sequence[QGraphicsItem] = []

        self._total_size: int = 0
        self._pressed: bool = False

        self.setAcceptHoverEvents(True)
        self._register_events()

    def refresh(self) -> None:
        self._gen_current_indicator()
        self._gen_checkpoint_indicators()
        self._gen_hover_indicator()

    def _register_events(self) -> None:
        self._dbg_watcher = DebuggerWatcher(self.on_debugger_state_updated, self.instance.debugger_mgr.debugger)
        self.on_debugger_state_updated()

    def on_debugger_state_updated(self) -> None:
        self._addr = 0
        self._total_size = 0
        self._checkpoints = []

        if isinstance(self._dbg_watcher.debugger.am_obj, BintraceDebugger):
            # FIXME: Expose trace info as TraceDebugger abstraction between Debugger<>BintraceDebugger
            dbg = self._dbg_watcher.debugger.am_obj
            t = dbg._trace.trace
            s = dbg._trace_dbg.state
            if s and s.event_count >= 0:
                self._addr = s.event_count
                self._total_size = t.get_num_events()
                self._checkpoints = [s.event_count for s in t.checkpoints]

        self.refresh()

    @property
    def width(self) -> int:
        return self._width

    def set_width(self, width: int) -> None:
        """
        Set the desired width of the trace map in scene units.
        """
        self.prepareGeometryChange()
        self._width = width

    @property
    def height(self) -> int:
        return self._height

    def set_height(self, height: int) -> None:
        """
        Set the desired height of the trace map in scene units.
        """
        self.prepareGeometryChange()
        self._height = height

    def paint(self, painter, option, widget) -> None:
        """
        Paint the trace map.
        """
        # Drawn by child items

    def boundingRect(self) -> QRectF:
        """
        Return the bounding dimensions of this item.
        """
        return QRectF(0, 0, self._width, self._height)

    def _get_pos_from_addr(self, addr: int) -> int | None:
        """
        Get scene X coordinate from address, or None if it could not be mapped.
        """
        if self._total_size == 0 or addr > self._total_size:
            return None
        return int(addr / self._total_size * self._width)

    def _get_addr_from_pos(self, pos: int) -> int | None:
        """
        Get address from scene X coordinate, or None if it could not be mapped.
        """
        offset = int(pos * self._total_size // self._width)
        return offset

    def _get_offset_size_rect(self, offset: int, size: int) -> QRectF:
        """
        Given a byte offset `offset` and number of bytes `size`, get a rect to draw.
        """
        if self._total_size == 0:
            return None
        x = offset / self._total_size * self._width
        width = size / self._total_size * self._width
        return QRectF(x, 0, width, self._height)

    def _create_line_indicator(
        self, addr: int, item_map, color=Qt.yellow, show_frontier: bool = False, z=None, z_frontier=None
    ) -> None:
        """
        Generate a cursor at a given address.
        """
        pos_x = self._get_pos_from_addr(addr)
        if pos_x is None:
            return

        pen = QPen(color)
        brush = QBrush(color)
        height = self.height

        tri_width = 7
        tri_height = 4

        pos_x = int(pos_x - tri_width / 2)  # Center drawing
        center = pos_x + int(tri_width / 2)
        pos_y = 0
        frontier_width = int(0.15 * max(self.width, self.height))

        if show_frontier:
            # Draw frontier gradients
            r = QRectF(center - frontier_width, pos_y, frontier_width, height)
            bg = QLinearGradient(r.topLeft(), r.topRight())
            color = Qt.red
            top_color = QColor(color)
            top_color.setAlpha(0)
            bg.setColorAt(0, top_color)
            bottom_color = QColor(color)
            bottom_color.setAlpha(180)
            bg.setColorAt(1, bottom_color)

            i = QGraphicsRectItem(r, parent=self)
            i.setPen(Qt.NoPen)
            i.setBrush(bg)
            if z_frontier is not None:
                i.setZValue(z_frontier)
            item_map.append(i)

            r = QRectF(center, pos_y, frontier_width, height)
            bg = QLinearGradient(r.topLeft(), r.topRight())
            color = Qt.blue
            top_color = QColor(color)
            bg.setColorAt(0, top_color)
            bottom_color = QColor(color)
            bottom_color.setAlpha(0)
            bg.setColorAt(1, bottom_color)

            i = QGraphicsRectItem(r, parent=self)
            i.setPen(Qt.NoPen)
            i.setBrush(bg)
            if z_frontier is not None:
                i.setZValue(z_frontier)
            item_map.append(i)

        # Draw line
        i = QGraphicsLineItem(center, 0, center, height, parent=self)
        i.setPen(pen)
        if z is not None:
            i.setZValue(z)
        item_map.append(i)

        # Draw top and bottom triangles
        t = QPolygonF()
        t.append(QPointF(pos_x, pos_y))
        t.append(QPointF(pos_x + tri_width - 1, pos_y))
        t.append(QPointF(center, pos_y + tri_height - 1))
        t.append(QPointF(pos_x, pos_y))

        pos_y += height - 1
        b = QPolygonF()
        b.append(QPointF(pos_x, pos_y))
        b.append(QPointF(center, pos_y - tri_height + 1))
        b.append(QPointF(pos_x + tri_width - 1, pos_y))
        b.append(QPointF(pos_x, pos_y))

        for i in [QGraphicsPolygonItem(t, parent=self), QGraphicsPolygonItem(b, parent=self)]:
            i.setPen(pen)
            i.setBrush(brush)
            if z is not None:
                i.setZValue(z)
            item_map.append(i)

    def _gen_checkpoint_indicators(self) -> None:
        """
        Create checkpoint indicators.
        """
        scene = self.scene()
        for item in self._checkpoint_items:
            scene.removeItem(item)
        self._checkpoint_items.clear()

        color = QColor(Qt.green)

        for checkpoint_addr in self._checkpoints:
            self._create_line_indicator(checkpoint_addr, self._checkpoint_items, color=color, z=self.ZVALUE_CHECKPOINT)

    def _gen_current_indicator(self) -> None:  # pylint: disable=unused-argument
        """
        Create current address indicator.
        """
        scene = self.scene()
        for item in self._indicator_items:
            scene.removeItem(item)
        self._indicator_items.clear()

        self._create_line_indicator(self._addr, self._indicator_items, show_frontier=True, z=self.ZVALUE_ADDR)

    def _gen_hover_indicator(self) -> None:
        """
        Create the hovered address indicator.
        """
        scene = self.scene()
        for item in self._hover_items:
            scene.removeItem(item)
        self._hover_items.clear()

        if self._hover_addr is not None:
            self._create_line_indicator(self._hover_addr, self._hover_items, color=Qt.gray, z=self.ZVALUE_HOVER)

    def _remove_hover_indicators(self) -> None:
        """
        Remove active hover items, if set.
        """
        if self._hover_addr is not None:
            self._hover_addr = None
            self._gen_hover_indicator()

    def mousePressEvent(self, event) -> None:
        if event.button() == Qt.LeftButton:
            pos = event.pos()
            offset = pos.x()
            self.select_offset(offset)
            self._pressed = True

    def mouseReleaseEvent(self, event) -> None:
        if event.button() == Qt.LeftButton:
            self._pressed = False

    def mouseMoveEvent(self, event) -> None:
        if self._pressed:
            pos = event.pos()
            offset = pos.x()
            self.select_offset(offset)
        else:
            super().mouseMoveEvent(event)

    def on_mouse_move_event_from_view(self, point: QPointF) -> None:
        """
        Add hover items.
        """
        self._remove_hover_indicators()
        self._hover_addr = self._get_addr_from_pos(point.x())
        self._gen_hover_indicator()

    def hoverLeaveEvent(self, event) -> None:  # pylint: disable=unused-argument
        self._remove_hover_indicators()

    def select_offset(self, offset: int) -> None:
        """
        Update listeners with new desired location.
        """
        if not isinstance(self._dbg_watcher.debugger.am_obj, BintraceDebugger):
            return

        addr = self._get_addr_from_pos(offset)
        if addr is None:
            return

        dbg = self._dbg_watcher.debugger.am_obj
        dbg.replay_to_nth_event(addr)


class QTraceMapView(QGraphicsView):
    """
    Graphics view for trace map scene. The scene will rotate based on dimensions of the viewport to support horizontal
    and vertical orientations.
    """

    def __init__(self, instance: Instance, parent=None) -> None:
        super().__init__(parent)
        self.instance = instance
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setMouseTracking(True)
        self._scene = QGraphicsScene(parent=self)
        self.setScene(self._scene)
        self.fm: TraceMapItem = TraceMapItem(self.instance)
        self._scale: float = 1.0
        self._scene.addItem(self.fm)
        self._orientation: str = "horizontal"
        self._base_width: int = 0

        self.setBackgroundBrush(Conf.palette_base)
        self.setResizeAnchor(QGraphicsView.NoAnchor)
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        self.update_size()

    def mouseMoveEvent(self, event) -> None:
        """
        Handle mouse move events.

        Mouse move events whilst not holding a mouse button will not be propagated down to QGraphicsItems, so we catch
        the movement events here in the view and forward them to the trace map item.
        """
        scene_pt = self.mapToScene(event.pos().x(), event.pos().y())
        item_pt = self.fm.mapFromScene(scene_pt)
        self.fm.on_mouse_move_event_from_view(item_pt)
        super().mouseMoveEvent(event)

    def wheelEvent(self, event) -> None:
        """
        Handle wheel events to scale and translate the trace map.
        """
        if event.modifiers() & Qt.ControlModifier == Qt.ControlModifier:
            self.adjust_viewport_scale(
                1.25 if event.angleDelta().y() > 0 else 1 / 1.25, QPoint(event.position().x(), event.position().y())
            )
        else:
            self.translate(100 * (-1 if event.angleDelta().y() < 0 else 1), 0)
            super().wheelEvent(event)

    def resizeEvent(self, event) -> None:  # pylint: disable=unused-argument
        """
        Handle view resize events, updating the trace map size accordingly.
        """
        self.update_size()

    def adjust_viewport_scale(self, scale: float | None = None, point: QPoint | None = None) -> None:
        """
        Adjust viewport scale factor.
        """
        if point is None:
            point = QPoint(0, 0)
        point_rel = self.mapToScene(point).x() / self.fm.width

        if scale is None:
            self._scale = 1.0
        else:
            self._scale = max(self._scale * scale, 1.0)

        self.update_size()
        self.translate(int(self.mapToScene(point).x() - point_rel * self.fm.width), 0)

    def keyPressEvent(self, event) -> None:
        """
        Handle key events.
        """
        if event.modifiers() & Qt.ControlModifier == Qt.ControlModifier:
            if event.key() == Qt.Key_0:
                self.adjust_viewport_scale()
                event.accept()
                return
            elif event.key() == Qt.Key_Equal:
                self.adjust_viewport_scale(1.25)
                event.accept()
                return
            elif event.key() == Qt.Key_Minus:
                self.adjust_viewport_scale(1 / 1.25)
                event.accept()
                return
        super().keyPressEvent(event)

    def changeEvent(self, event: QEvent) -> None:
        """
        Redraw on color scheme update.
        """
        if event.type() == QEvent.StyleChange:
            self.setBackgroundBrush(Conf.palette_base)
            self.fm.refresh()

    def update_size(self) -> None:
        """
        Resize map.
        """
        rotation = 0
        vg = self.viewport().geometry()
        if vg.width() > vg.height():
            if self._orientation != "horizontal":
                rotation = -90
            self._orientation = "horizontal"
            w, h = vg.width(), vg.height()
        else:
            if self._orientation != "vertical":
                rotation = 90
            self._orientation = "vertical"
            w, h = vg.height(), vg.width()

        if rotation:
            self._scale = 1.0

        if self._scale <= 1.0:
            # Only resize to map to viewport width if scale is at base level to not disturb preferred size
            self._base_width = w

        self.fm.set_width(int(self._base_width * self._scale))
        self.fm.set_height(h)
        self.fm.refresh()
        self.setSceneRect(QRectF(0, 0, self.fm.width, self.fm.height))
        if rotation:
            self.rotate(rotation)


class QTraceMap(QWidget):
    """
    Map of the current trace, with debugger playback position and checkpoint indicators.
    """

    def __init__(self, instance: Instance, parent=None) -> None:
        super().__init__(parent)
        self.instance = instance
        self.view: QTraceMapView = None
        self._init_widgets()

    def sizeHint(self):  # pylint:disable=no-self-use
        return QSize(25, 25)

    def minimumSizeHint(self):  # pylint:disable=no-self-use
        return QSize(25, 25)

    #
    # Public methods
    #

    def refresh(self) -> None:
        if self.view is not None:
            self.view.fm.refresh()

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self.view = QTraceMapView(self.instance, self)
        layout = QHBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)



from __future__ import annotations

from angrmanagement.logic import GlobalInfo

from .job import Job


class DecompileFunctionJob(Job):
    """
    The job for running the decompiler analysis. You can trigger this by pressing f5 in a function.
    """

    def __init__(self, function, on_finish=None, blocking: bool = False, **kwargs) -> None:
        self.kwargs = kwargs
        self.function = function
        super().__init__(name="Decompiling", on_finish=on_finish, blocking=blocking)

    def _run(self, inst) -> None:
        decompiler = inst.project.analyses.Decompiler(
            self.function,
            flavor="pseudocode",
            variable_kb=inst.pseudocode_variable_kb,
            **self.kwargs,
            progress_callback=self._progress_callback,
        )
        # cache the result
        inst.kb.structured_code[(self.function.addr, "pseudocode")] = decompiler.cache

        GlobalInfo.main_window.workspace.plugins.decompile_callback(self.function)



from __future__ import annotations

from .job import Job


class VFGGenerationJob(Job):
    def __init__(self, addr: int) -> None:
        super().__init__("VFG generation")
        self._addr = addr

    def _run(self, inst):
        return inst.project.analyses.VFG(function_start=self._addr)

    def finish(self, inst, result) -> None:
        super().finish(inst, result)
        inst.vfgs[self._addr] = result

    def __repr__(self) -> str:
        return f"Generating VFG for function at {self._addr:#x}"



from __future__ import annotations

import tomlkit
from tomlkit.items import AoT, Integer, String, Table


class PluginDescription:
    """
    Describes an angr management plugin. Can be generated from plugin.toml.
    """

    def __init__(self) -> None:
        # Metadata
        self.plugin_metadata_version: int = None

        # Plugin
        self.name: str = ""
        self.shortname: str = ""
        self.version: str = ""
        self.description: str = ""
        self.long_description: str = ""
        self.platforms: list[str] = []
        self.min_angr_vesion: str = ""
        self.author = ""
        self.entrypoints: list[str] = []
        self.require_workspace: bool = True
        self.has_url_actions: bool = False

        # file path
        self.plugin_file_path: str = ""

    @classmethod
    def load_single_plugin(cls, data: Table) -> PluginDescription:
        desc = PluginDescription()

        desc.name = data.get("name", None)
        if not isinstance(desc.name, String):
            raise TypeError(f'"name" must be a String instance, not a {type(desc.name)}')
        if not desc.name:
            raise TypeError('"name" cannot be empty')

        desc.shortname = data.get("shortname", None)
        if not isinstance(desc.shortname, String):
            raise TypeError(f'"shortname" must be a String instance, not a {type(desc.shortname)}')
        if not desc.shortname:
            raise TypeError('"shortname" cannot be empty')

        desc.entrypoints = data.get("entrypoints", "")
        if not isinstance(desc.entrypoints, list) or not all(
            isinstance(entrypoint, String) for entrypoint in desc.entrypoints
        ):
            raise TypeError('"entrypoints" must be a List of String instances')
        if not desc.entrypoints:
            raise TypeError('"entrypoints" cannot be empty')

        # optional
        desc.version = data.get("version", "")
        desc.description = data.get("description", "")
        desc.long_description = data.get("long_description", "")
        desc.platforms = data.get("platforms", "")
        desc.min_angr_vesion = data.get("min_angr_version", "")
        desc.author = data.get("author", "")
        desc.require_workspace = data.get("require_workspace", True)
        desc.has_url_actions = data.get("has_url_actions", False)

        return desc

    @classmethod
    def from_toml(cls, file_path: str) -> list[PluginDescription]:
        with open(file_path, encoding="utf-8") as f:
            data = tomlkit.load(f)

        # load metadata
        outer_desc = PluginDescription()
        if (
            "meta" in data
            and "plugin_metadata_version" in data["meta"]
            and isinstance(data["meta"]["plugin_metadata_version"], Integer)
        ):
            outer_desc.plugin_metadata_version = data["meta"]["plugin_metadata_version"].unwrap()

        if outer_desc.plugin_metadata_version is None:
            raise TypeError("Cannot find plugin_metadata_version")
        if outer_desc.plugin_metadata_version != 0:
            raise TypeError(f"Unsupported plugin metadata version {outer_desc.plugin_metadata_version}")

        descs = []
        # load plugin information
        if "plugins" in data and isinstance(data["plugins"], AoT):
            # multiple plugins to load!
            for plugin in data["plugins"]:
                desc = PluginDescription.load_single_plugin(plugin)
                desc.plugin_metadata_version = outer_desc.plugin_metadata_version
                desc.plugin_file_path = file_path
                descs.append(desc)
        elif "plugin" in data:
            desc = PluginDescription.load_single_plugin(data["plugin"])
            desc.plugin_metadata_version = outer_desc.plugin_metadata_version
            desc.plugin_file_path = file_path
            descs.append(desc)
        else:
            raise TypeError('Cannot find any "plugin" or "plugins" table.')

        return descs



from __future__ import annotations

import multiprocessing
import platform
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    import angr

    from angrmanagement.data.instance import Instance


def extract_first_paragraph_from_docstring(desc: str) -> str:
    desc = desc.splitlines()
    last_line, first_line = -1, -1
    for idx, line in enumerate(desc):
        if first_line < 0:
            if len(line.strip()) > 0:
                first_line = idx
        else:
            if len(line.strip()) == 0:
                last_line = idx
                break

    if first_line >= 0:
        if last_line < 0:
            last_line = len(desc)
        desc = desc[first_line:last_line]
        num_whitespace_chars = len(desc[0]) - len(desc[0].lstrip())
        desc = " ".join(line[num_whitespace_chars:] for line in desc)
    else:
        desc = ""

    return desc


class AnalysesConfiguration:
    """
    Configuration for a sequence of analyses.
    """

    def __init__(self, analyses: Sequence[AnalysisConfiguration], instance: Instance) -> None:
        self.instance = instance
        self.analyses: Sequence[AnalysisConfiguration] = analyses

    def __len__(self) -> int:
        return len(self.analyses)

    def __iter__(self):
        return iter(self.analyses)

    def __getitem__(self, key: int | str):
        if isinstance(key, int):
            return self.analyses[key]
        return self.by_name(key)

    def by_name(self, name: str) -> AnalysisConfiguration:
        for a in self.analyses:
            if a.name == name:
                return a
        raise KeyError(name)


class AnalysisConfiguration:
    """
    Configuration for an analysis.
    """

    def __init__(self, instance: Instance) -> None:
        self.instance = instance
        self.project: angr.Project = self.instance.project.am_obj
        self.enabled: bool = False
        self.name: str = ""
        self.display_name: str = ""
        self.description: str = "Description not available"
        self.options: Mapping[str, AnalysisOption] = {}

    def __getitem__(self, key: str):
        return self.options[key]

    def to_dict(self):
        """
        Return dictionary with configuration for this option.
        """
        o = {}
        self.update_dict(o)
        return o

    def update_dict(self, out: Mapping[str, Any]) -> None:
        """
        Update dictionary `out` with configuration for this option.
        """
        for o in self.options.values():
            if getattr(o, "enabled", None) is False:
                continue
            o.update_dict(out)


class AnalysisOption:
    """
    Configurable option for an analysis.
    """

    def __init__(self, name: str, display_name: str, tooltip: str) -> None:
        self.name: str = name
        self.display_name: str = display_name
        self.tooltip: str = tooltip

    def update_dict(self, out: Mapping[str, Any]) -> None:
        """
        Update dictionary `out` with configuration for this option.
        """


class PrimitiveAnalysisOption(AnalysisOption):
    """
    Configurable option for an analysis, with a fundamental type (e.g. bool)
    """

    def __init__(self, name: str, description: str, default: Any, tooltip: str) -> None:
        super().__init__(name, description, tooltip)
        self.default: Any = default
        self.value: Any = default

    def update_dict(self, out: Mapping[str, Any]) -> None:
        """
        Update `out` dictionary with configuration for this option.
        """
        out[self.name] = self.value


class BoolAnalysisOption(PrimitiveAnalysisOption):
    """
    Boolean option for an analysis.
    """

    def __init__(self, name: str, description: str, default: bool = False, tooltip: str = "") -> None:
        super().__init__(name, description, default, tooltip)


class StringAnalysisOption(PrimitiveAnalysisOption):
    """
    String option for an analysis.

    :ivar optional: If this option is optional or mandatory.
    :ivar enabled:  Is this option enabled by the user or not.
    """

    def __init__(
        self, name: str, description: str, default: str = "", tooltip: str = "", optional: bool = False
    ) -> None:
        self.optional = optional
        self.enabled = not self.optional
        super().__init__(name, description, default, tooltip)


class IntAnalysisOption(PrimitiveAnalysisOption):
    """
    Integer option for an analysis.
    """

    def __init__(
        self,
        name: str,
        description: str,
        default: int = 0,
        tooltip: str = "",
        minimum: int | None = None,
        maximum: int | None = None,
    ) -> None:
        super().__init__(name, description, default, tooltip)
        self.minimum_value = minimum
        self.maximum_value = maximum


class ChoiceAnalysisOption(PrimitiveAnalysisOption):
    """
    A multi-value choice.
    """

    def __init__(
        self, name: str, description: str, choices: Mapping[Any, str], default: Any, tooltip: str = ""
    ) -> None:
        super().__init__(name, description, default, tooltip)
        self.choices = choices


class CFGForceScanMode(Enum):
    """
    CFG scanning mode options.
    """

    Disabled = 0
    SmartScan = 1
    CompleteScan = 2


class CFGAnalysisConfiguration(AnalysisConfiguration):
    """
    Configuration for CFGFast analysis.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "cfg"
        self.display_name = "Control-Flow Graph Recovery"
        self.description = extract_first_paragraph_from_docstring(self.project.analyses.CFGFast.__doc__)
        self.enabled = True
        self.options = {
            o.name: o
            for o in [
                BoolAnalysisOption("resolve_indirect_jumps", "Resolve indirect jumps", True),
                BoolAnalysisOption("data_references", "Collect cross-references and guess data types", True),
                BoolAnalysisOption("cross_references", "Perform deep analysis on cross-references (slow)"),
                BoolAnalysisOption("skip_unmapped_addrs", "Skip unmapped addresses", True),
                BoolAnalysisOption("exclude_sparse_regions", "Exclude Sparse Regions", True),
                BoolAnalysisOption(
                    "explicit_analysis_starts", "Exclude non-explicit functions for analysis (incomplete)", False
                ),
                ChoiceAnalysisOption(
                    "scanning_mode",
                    "Scan to maximize identified code blocks",
                    {
                        CFGForceScanMode.Disabled: "Disabled",
                        CFGForceScanMode.SmartScan: "Smart Scan",
                        CFGForceScanMode.CompleteScan: "Complete Scan",
                    },
                    CFGForceScanMode.SmartScan,
                ),
                StringAnalysisOption(
                    "regions",
                    "Regions for analysis",
                    tooltip="Specify ranges of regions for which to recover CFG. Example: 0x400000-0x401000. You may "
                    "specify multiple address ranges for analysis.",
                    optional=True,
                ),
                StringAnalysisOption(
                    "function_starts",
                    "Start at function addresses",
                    tooltip="Specify function addresses to start recursive descent of CFG generation to speed up "
                    "analysis. Example: 0x400000,0x401000",
                    optional=True,
                ),
            ]
        }


class FlirtAnalysisConfiguration(AnalysisConfiguration):
    """
    Configuration for Flirt analysis.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "flirt"
        self.display_name = "Function Signature Matching"
        self.description = self.project.analyses.Flirt.__doc__.strip()
        self.enabled = True


class CodeTaggingConfiguration(AnalysisConfiguration):
    """
    Configuration for Code Tagging.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "code_tagging"
        self.display_name = "Tag Functions Based on Syntactic Features"
        self.description = "Add tags to functions based on syntactic features in assembly code and referenced strings."
        self.enabled = False


class VariableRecoveryConfiguration(AnalysisConfiguration):
    """
    Configuration for VariableRecovery analysis.
    """

    SMALL_BINARY_SIZE = 65536
    MEDIUM_BINARY_SIZE = 400000

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "varec"
        self.display_name = "Recover Variables on All Functions"
        self.description = (
            "Perform a full-project variable recovery and calling-convention recovery analysis. "
            "Recommended for small- to medium-sized binaries. This analysis takes a long time to "
            "finish on large binaries. You can manually perform a variable recovery and "
            "calling-convention recovery analysis on an individual basis after loading the project."
        )
        self.enabled = self.get_main_obj_size() <= self.MEDIUM_BINARY_SIZE
        self.options = {
            o.name: o
            for o in [
                IntAnalysisOption(
                    "workers",
                    "Number of parallel workers",
                    tooltip="0 to disable parallel analysis. Default to the number of available cores "
                    "minus one in the local system. Automatically default to 0 for small binaries "
                    "on all platforms, and small- to medium-sized binaries on Windows and MacOS "
                    "(to avoid the cost of spawning new angr-management processes).",
                    default=self.get_default_workers(),
                    minimum=0,
                ),
                BoolAnalysisOption(
                    "skip_signature_matched_functions",
                    "Skip variable recovery for signature-matched functions",
                    True,
                ),
            ]
        }

    def get_main_obj_size(self) -> int:
        main_obj_size = 0
        if self.project.loader.main_object is not None:
            main_obj = self.project.loader.main_object
            if main_obj.segments:
                for seg in main_obj.segments:
                    if seg.is_executable:
                        main_obj_size += seg.memsize
            if main_obj_size == 0 and main_obj.sections:
                # fall back to sections
                for sec in main_obj.sections:
                    if sec.is_executable:
                        main_obj_size += sec.memsize
        return main_obj_size

    def get_default_workers(self) -> int:
        main_obj_size = self.get_main_obj_size()

        default_workers = max(multiprocessing.cpu_count() - 1, 1)
        if default_workers == 1:
            return 0

        if platform.system() in {"Windows", "Darwin"}:
            if main_obj_size <= self.MEDIUM_BINARY_SIZE:
                return 0
            return default_workers

        if main_obj_size <= self.SMALL_BINARY_SIZE:
            return 0
        return default_workers



from __future__ import annotations

import functools
import logging
from collections.abc import Callable, Sequence
from enum import Enum
from typing import TYPE_CHECKING

import angr
import PySide6
from angr import Block
from angr.knowledge_plugins.cfg import MemoryData, MemoryDataSort
from angr.knowledge_plugins.patches import Patch
from PySide6.QtCore import QEvent, QMarginsF, QPointF, QRectF, QSizeF, Qt, QTimer, Signal
from PySide6.QtGui import QAction, QColor, QCursor, QFont, QPainterPath, QPen, QWheelEvent
from PySide6.QtWidgets import (
    QAbstractScrollArea,
    QAbstractSlider,
    QApplication,
    QComboBox,
    QFrame,
    QGraphicsItem,
    QGraphicsObject,
    QGraphicsScene,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
    QGraphicsView,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
)

from angrmanagement.config import Conf
from angrmanagement.data.breakpoint import Breakpoint, BreakpointType
from angrmanagement.logic.debugger import DebuggerWatcher
from angrmanagement.ui.dialogs.input_prompt import InputPromptDialog
from angrmanagement.ui.dialogs.jumpto import JumpTo
from angrmanagement.utils import is_printable

from .view import SynchronizedInstanceView

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace

log = logging.getLogger(__name__)

RowCol = tuple[int, int]
HexByteValue = int | str
HexAddress = int
HexDataBuffer = bytes | bytearray
HexDataProvider = Callable[[HexAddress], HexByteValue]


class HexDataSource(Enum):
    """
    Data source to be displayed in the hex view.
    """

    Loader = 0
    Debugger = 1


class HexHighlightRegion:
    """
    Defines a highlighted region.
    """

    def __init__(self, color: QColor, addr: HexAddress, size: int, tooltip: str | None = None) -> None:
        self.color: QColor = color
        self.addr: HexAddress = addr
        self.size: int = size
        self.active: bool = False
        self._tooltip: str | None = tooltip

    def gen_context_menu_actions(self) -> QMenu | None:  # pylint: disable=no-self-use
        """
        Get submenu for this highlight region.
        """
        return None

    def get_tooltip(self) -> str | None:
        """
        Return a tooltip for this region.
        """
        return self._tooltip


class BreakpointHighlightRegion(HexHighlightRegion):
    """
    Defines a highlighted region indicating a patch.
    """

    def __init__(self, bp: Breakpoint, view: HexView) -> None:
        super().__init__(Qt.cyan, bp.addr, bp.size)
        self.bp: Breakpoint = bp
        self.view: HexView = view

    def gen_context_menu_actions(self) -> QMenu | None:
        """
        Get submenu for this highlight region.
        """
        bp_type_str = {
            BreakpointType.Execute: "Execute",
            BreakpointType.Read: "Read",
            BreakpointType.Write: "Write",
        }.get(self.bp.type)
        mnu = QMenu(f"Breakpoint 0x{self.bp.addr:x} {bp_type_str} ({self.bp.size} bytes)")
        act = QAction("&Remove", mnu)
        act.triggered.connect(self.remove)
        mnu.addAction(act)
        return mnu

    def remove(self) -> None:
        """
        Remove this breakpoint.
        """
        self.view.instance.breakpoint_mgr.remove_breakpoint(self.bp)

    def get_tooltip(self) -> str | None:
        """
        Return a tooltip for this region.
        """
        bp_type_str = {
            BreakpointType.Execute: "Execute",
            BreakpointType.Read: "Read",
            BreakpointType.Write: "Write",
        }.get(self.bp.type)
        return f"Breakpoint 0x{self.bp.addr:x} {bp_type_str} ({self.bp.size} bytes)"


class PatchHighlightRegion(HexHighlightRegion):
    """
    Defines a highlighted region indicating a patch.
    """

    def __init__(self, patch: Patch, view: HexView) -> None:
        super().__init__(Qt.yellow, patch.addr, len(patch))
        self.patch: Patch = patch
        self.view: HexView = view

    def get_tooltip(self) -> str | None:
        """
        Return a tooltip for this region.
        """
        return f"Patch 0x{self.patch.addr:x} ({len(self.patch)} bytes)"

    def gen_context_menu_actions(self) -> QMenu | None:
        """
        Get submenu for this highlight region.
        """
        mnu = QMenu(f"Patch 0x{self.patch.addr:x} ({len(self.patch)} bytes)")
        act = QAction("&Split", mnu)
        act.triggered.connect(self.split)
        act.setEnabled(self.can_split())
        mnu.addAction(act)
        act = QAction("Set &Comment...", mnu)
        act.triggered.connect(self.comment)
        mnu.addAction(act)
        mnu.addSeparator()
        act = QAction("&Revert", mnu)
        act.triggered.connect(self.revert_with_prompt)
        mnu.addAction(act)
        return mnu

    def revert_with_prompt(self) -> None:
        """
        Revert this patch. Prompt for confirmation.
        """
        dlg = QMessageBox()
        dlg.setWindowTitle("Revert patch")
        dlg.setText("Are you sure you want to revert this patch?")
        dlg.setIcon(QMessageBox.Question)
        dlg.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        dlg.setDefaultButton(QMessageBox.Cancel)
        if dlg.exec_() != QMessageBox.Yes:
            return
        self.revert()

    def revert(self) -> None:
        """
        Revert this patch.
        """
        pm = self.view.instance.patches
        pm.remove_patch(self.patch.addr)
        pm.am_event()

    def can_split(self) -> bool:
        """
        Determine if this patch can be split based on current cursor location.
        """
        cursor = self.view.inner_widget.hex.cursor
        return self.patch.addr < cursor < (self.patch.addr + len(self.patch))

    def split(self) -> None:
        """
        Split this patch at view's current cursor location.
        """
        cursor = self.view.inner_widget.hex.cursor
        if self.can_split():
            o = cursor - self.patch.addr
            new_patch = Patch(cursor, self.patch.new_bytes[o:], self.patch.comment)
            self.patch.new_bytes = self.patch.new_bytes[0:o]
            pm = self.view.instance.patches
            pm.add_patch_obj(new_patch)
            pm.am_event()

    def can_merge_with(self, other: PatchHighlightRegion) -> bool:
        """
        Determine if this patch can be merged with `other`. We only consider directly adjacent patches.
        """
        return other.patch.addr == (self.patch.addr + len(self.patch))

    def merge_with(self, other: PatchHighlightRegion) -> None:
        """
        Merge `other` into this patch.
        """
        if self.can_merge_with(other):
            self.patch.new_bytes += other.patch.new_bytes
            pm = self.view.instance.patches
            pm.remove_patch(other.patch.addr)
            pm.am_event()

    def comment(self) -> None:
        """
        Set the comment for this patch.
        """
        dlg = InputPromptDialog("Set Patch Comment", "Patch comment:", self.patch.comment, parent=self.view)
        dlg.exec_()
        if dlg.result:
            self.patch.comment = dlg.result
            pm = self.view.instance.patches
            pm.am_event()


class HexGraphicsObject(QGraphicsObject):
    """
    A graphics item providing a conventional hex-editor interface for a contiguous region of memory.
    """

    cursor_changed = Signal()
    viewport_changed = Signal()

    def __init__(self) -> None:
        super().__init__()
        self.setFlag(QGraphicsItem.ItemUsesExtendedStyleOption, True)  # Give me more specific paint update rect info
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)  # Give me focus/key events
        self.setFlag(QGraphicsItem.ItemClipsToShape, True)
        self.display_offset_addr: HexAddress = 0
        self.display_num_rows: int = 1
        self.display_start_addr: HexAddress = 0
        self.display_end_addr: HexAddress = 0
        self.start_addr: HexAddress = 0
        self.num_bytes: int = 0
        self.end_addr: HexAddress = 0  # Exclusive
        self.read_func: HexDataProvider | None = None
        self.write_func: Callable[[HexAddress, HexByteValue], bool] = None
        self.data: HexDataBuffer | None = None
        self.addr_offset: int = 0
        self.addr_width: int = 0
        self.ascii_column_offsets: Sequence[int] = []
        self.ascii_space: int = 0
        self.ascii_width: int = 0
        self.byte_column_offsets: Sequence[int] = []
        self.byte_group_space: int = 0
        self.byte_space: int = 0
        self.byte_width: int = 0
        self.char_width: int = 0
        self.max_x: int = 0
        self.max_y: int = 0
        self.row_height: int = 0
        self.row_padding: int = 0
        self.section_space: int = 0
        self.cursor: HexAddress = self.start_addr
        self.cursor_nibble: int | None = None
        self.selection_start: HexAddress | None = None
        self.mouse_pressed: bool = False
        self.num_rows: int = 0
        self.font: QFont = QFont(Conf.disasm_font)
        self.font.setPointSizeF(12)
        self.ascii_column_active: bool = False
        self.cursor_blink_state: bool = True
        self.highlighted_regions: Sequence[HexHighlightRegion] = []

        self.cursor_blink_timer: QTimer = QTimer(self)
        self.cursor_blink_timer.timeout.connect(self.toggle_cursor_blink)
        self.show_cursor: bool = True
        self._processing_cursor_update: bool = False
        self.always_show_cursor: bool = False

        self._update_layout()

    def focusInEvent(self, event: PySide6.QtGui.QFocusEvent) -> None:  # pylint: disable=unused-argument
        """
        Item receives focus.
        """
        self.show_cursor = True
        self.restart_cursor_blink_timer()
        self.update()

    def focusOutEvent(self, event: PySide6.QtGui.QFocusEvent) -> None:  # pylint: disable=unused-argument
        """
        Item lost focus.
        """
        self.cursor_blink_timer.stop()
        self.show_cursor = self.always_show_cursor
        self.cursor_blink_state = self.always_show_cursor
        self.update()

    def set_always_show_cursor(self, always_show: bool) -> None:
        """
        Set policy of whether the cursor should always be shown (when focus is lost) or not.
        """
        self.always_show_cursor = always_show
        if not self.cursor_blink_timer.isActive():
            self.show_cursor = self.always_show_cursor
            self.cursor_blink_state = self.always_show_cursor
            self.update()

    def _set_display_common(self) -> None:
        """
        Handle common update of offset or row count changed.
        """
        self.display_end_addr = min(self.end_addr, self.display_start_addr + self.display_num_rows * 16)
        self._update_layout()
        self.viewport_changed.emit()

    def set_display_num_rows(self, num_rows: int) -> None:
        """
        Set number of rows to display.
        """
        self.display_num_rows = max(num_rows, 1)
        self._set_display_common()

    def set_display_offset(self, offset: HexAddress) -> None:
        """
        Set displayed range offset.
        """
        self.display_offset_addr = max(0, min(offset & ~0xF, self.end_addr - self.start_addr - 0x10))
        self.display_start_addr = self.start_addr + self.display_offset_addr
        self._set_display_common()

    def move_viewport_to(self, addr: HexAddress, preserve_relative_offset: bool = False) -> None:
        """
        Translate the visible range so `addr` is visible.
        """
        assert self.start_addr <= addr < self.end_addr

        if preserve_relative_offset and (self.display_start_addr <= self.cursor < self.display_end_addr):
            # Let the target addr be on the same relative row from top of screen
            self.set_display_offset(addr - self.cursor + self.display_start_addr - self.start_addr)
            return

        if addr < self.display_start_addr:
            # Let the target addr be on the first displayed row
            self.set_display_offset(addr - self.start_addr)
            return

        # Let the target addr be on last fully displayed row
        max_fully_visible_bytes = 0x10 * (self.display_num_rows - 1)
        display_end = self.display_start_addr + max_fully_visible_bytes
        if addr >= display_end:
            offset = addr - max_fully_visible_bytes - self.start_addr + 0x10
            self.set_display_offset(offset)

    def _set_data_common(self) -> None:
        """
        Common handler for set_data_*
        """
        assert self.num_bytes >= 0
        self.num_rows = int((self.num_bytes + (self.start_addr & 0xF) + 0xF) / 16)
        self.end_addr = self.start_addr + self.num_bytes
        self.clear_selection()
        self.set_display_offset(0)
        self.set_cursor(self.start_addr)
        self._update_layout()

    def _simple_read_callback(self, addr: HexAddress) -> HexByteValue:
        """
        Handler for simple data buffers.
        """
        return self.data[addr - self.start_addr]

    # pylint:disable=unused-argument,no-self-use
    def _simple_write_callback(self, addr: HexAddress, value: HexByteValue) -> bool:
        """
        Handler for simple data buffers.
        """
        return False

    def set_data(self, data: HexDataBuffer, start_addr: HexAddress = 0, num_bytes: int | None = None) -> None:
        """
        Assign the buffer to be displayed with bytes.
        """
        self.data = data
        self.start_addr = start_addr
        self.num_bytes = num_bytes if num_bytes is not None else len(data)
        self.read_func = self._simple_read_callback
        self.write_func = self._simple_write_callback
        self._set_data_common()

    def set_data_callback(self, write_func, read_func: HexDataProvider, start_addr: HexAddress, num_bytes: int) -> None:
        """
        Assign the buffer to be displayed using a callback function.
        """
        self.data = None
        self.start_addr = start_addr
        self.num_bytes = num_bytes
        self.write_func = write_func
        self.read_func = read_func
        self._set_data_common()

    def clear(self) -> None:
        """
        Clear the current buffer.
        """
        self.set_data(b"")

    def point_to_row(self, p: QPointF) -> int | None:
        """
        Return index of row containing point `p`, or None if the point is not contained.
        """
        row = int(p.y() / self.row_height)
        return row if row < self.num_rows else None

    @staticmethod
    def point_to_column(p: QPointF, columns: Sequence[int]) -> int | None:
        """
        Given a point `p` and list of column offsets `columns`, return the index of column point p or None if the point
        is not contained.
        """
        x = p.x()
        for i in range(len(columns) - 1):
            if columns[i] <= x < columns[i + 1]:
                return i
        return None

    def point_to_addr(self, pt: QPointF) -> tuple[int, bool] | None:
        """
        Get the (address, ascii_column) tuple for a given point. If the point falls within the bytes display region,
        ascii_column is False. If the point falls within the ASCII display region, ascii_column is True.
        """
        ascii_column = False
        row = self.point_to_row(pt)
        if row is None:
            return None
        col = self.point_to_column(pt, self.byte_column_offsets)
        if col is None:
            col = self.point_to_column(pt, self.ascii_column_offsets)
            if col is None:
                return None
            else:
                ascii_column = True
        addr = self.row_col_to_addr(row, col)
        if addr < self.start_addr or addr >= self.end_addr:
            return None
        return addr, ascii_column

    def row_to_point(self, row: int) -> QPointF:
        """
        Get a point in the scene for a given row.
        """
        return QPointF(0, row * self.row_height)

    def row_col_to_point(self, row: int, col: int, ascii_section: bool = False) -> QPointF:
        """
        Get point for (row, col) in either ASCII section or bytes section.
        """
        columns = self.ascii_column_offsets if ascii_section else self.byte_column_offsets
        return QPointF(columns[col], self.row_to_point(row).y())

    def addr_to_point(self, addr: HexAddress, ascii_section: bool = False) -> QPointF:
        """
        Get point for address `addr` in either ASCII section or bytes section.
        """
        return self.row_col_to_point(*self.addr_to_row_col(addr), ascii_section)

    def addr_to_rect(self, addr: HexAddress) -> QRectF:
        """
        Get rect for address `addr` in whichever section is currently active.
        """
        if self.ascii_column_active:
            column_width = self.ascii_width
            column_space = self.ascii_space
        else:
            column_width = self.byte_width
            column_space = self.byte_space

        pt = self.addr_to_point(addr, self.ascii_column_active)
        pt.setX(pt.x() - column_space / 2)
        return QRectF(pt, QSizeF(column_width + column_space, self.row_height))

    def row_to_addr(self, row: int) -> int:
        """
        Get address for a given row.
        """
        return (self.display_start_addr & ~15) + row * 16

    def row_col_to_addr(self, row: int, col: int) -> int:
        """
        Get address for a given row, column.
        """
        return (self.display_start_addr & ~15) + row * 16 + col

    def addr_to_row_col(self, addr: int) -> RowCol:
        """
        Get (row, column) for a given address.
        """
        addr = addr - (self.display_start_addr & ~0xF)
        row = addr >> 4
        col = addr & 15
        return row, col

    def begin_selection(self) -> None:
        """
        Begin selection at current cursor.
        """
        self.selection_start = self.cursor
        self.update()

    def clear_selection(self) -> None:
        """
        Clear selection.
        """
        self.selection_start = None
        self.update()

    def restart_cursor_blink_timer(self) -> None:
        """
        Restart the cursor blink timer.
        """
        self.cursor_blink_timer.stop()
        self.cursor_blink_state = True
        self.cursor_blink_timer.start(750)

    def update_active_highlight_regions(self) -> None:
        """
        Update active property on highlight regions.
        """
        if self.selection_start is None:
            minaddr = self.cursor
            maxaddr = self.cursor
        else:
            minaddr = min(self.cursor, self.selection_start)
            maxaddr = max(self.cursor, self.selection_start)

        for region in self.highlighted_regions:
            region_end = region.addr + region.size - 1
            region.active = not (region.addr > maxaddr or minaddr > region_end)

    def get_active_highlight_regions(self) -> Sequence[HexHighlightRegion]:
        """
        Get currently active highlight regions.
        """
        return [rgn for rgn in self.highlighted_regions if rgn.active]

    def get_highlight_regions_at_addr(self, addr: HexAddress) -> Sequence[HexHighlightRegion]:
        """
        Return the highlight regions at specified address.
        """
        regions = []
        for region in self.highlighted_regions:
            if region.addr <= addr < (region.addr + region.size):
                regions.append(region)
        return regions

    def get_highlight_regions_under_cursor(self) -> Sequence[HexHighlightRegion]:
        """
        Return the highlight regions under the cursor.
        """
        return self.get_highlight_regions_at_addr(self.cursor)

    def set_cursor(
        self, addr: int, ascii_column: bool | None = None, nibble: int | None = None, update_viewport: bool = True
    ) -> None:
        """
        Move cursor to address `addr`.
        """
        if addr >= self.end_addr or addr < self.start_addr:
            return
        if self._processing_cursor_update:
            return
        self._processing_cursor_update = True
        if ascii_column is not None:
            self.ascii_column_active = ascii_column
        if self.hasFocus():
            self.restart_cursor_blink_timer()
        cursor_changed = self.cursor != addr
        if cursor_changed:
            if update_viewport:
                self.move_viewport_to(addr)
            self.cursor = addr
        self.cursor_nibble = nibble
        self.cursor_changed.emit()
        self.update_active_highlight_regions()
        self.update()
        self._processing_cursor_update = False

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handle mouse press events (e.g. updating selection).
        """
        if event.button() == Qt.LeftButton:
            addr = self.point_to_addr(event.pos())
            if addr is None:
                return
            addr, ascii_column = addr
            self.mouse_pressed = True
            if QApplication.keyboardModifiers() in (Qt.ShiftModifier,):
                if self.selection_start is None:
                    self.begin_selection()
            else:
                self.clear_selection()
            self.set_cursor(addr, ascii_column)
            event.accept()

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handle mouse double-click events (e.g. update selection)
        """
        if event.button() == Qt.LeftButton:
            regions = sorted(self.get_highlight_regions_under_cursor(), key=lambda r: self.cursor - r.addr)
            if len(regions) > 0:
                region = regions[0]
                self.set_cursor(region.addr + region.size - 1)
                self.begin_selection()
                self.set_cursor(region.addr)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handle mouse move events (e.g. updating selection).
        """
        if self.mouse_pressed:
            addr = self.point_to_addr(event.pos())
            if addr is None:
                return
            addr, ascii_column = addr
            if self.selection_start is None:
                self.begin_selection()
            self.set_cursor(addr, ascii_column)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handle mouse release events.
        """
        if event.button() == Qt.LeftButton:
            self.mouse_pressed = False

    def _set_byte_value(self, value: int) -> None:
        """
        Handle byte modification via user entry.
        """
        if not self.write_func(self.cursor, value):
            return
        self.set_cursor(self.cursor + 1)

    def _set_nibble_value(self, value: int) -> None:
        nibble = 1 if (self.cursor_nibble is None) else self.cursor_nibble
        new_byte_value = self.read_func(self.cursor)
        if not isinstance(new_byte_value, int):
            new_byte_value = 0
        else:
            new_byte_value &= ~(0xF << (nibble * 4))
        new_byte_value |= value << (nibble * 4)
        if not self.write_func(self.cursor, new_byte_value):
            return
        next_nibble = 1 - nibble
        self.set_cursor(self.cursor + next_nibble, nibble=next_nibble)

    def keyPressEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        """
        Handle key press events (e.g. moving cursor around).
        """
        movement_keys = {
            Qt.Key_Up: -16,
            Qt.Key_Down: 16,
            Qt.Key_Right: 1,
            Qt.Key_Left: -1,
            Qt.Key_PageUp: 0,
            Qt.Key_PageDown: 0,
            Qt.Key_Home: 0,
            Qt.Key_End: 0,
        }
        if event.key() in movement_keys:
            if QApplication.keyboardModifiers() & Qt.ShiftModifier:
                if self.selection_start is None:
                    self.begin_selection()
            else:
                self.clear_selection()

            # FIXME: When holding Ctrl, only scroll the viewport
            preserve_relative_offset = False
            if event.key() == Qt.Key_PageUp:
                new_cursor = max(self.start_addr, self.cursor - (self.display_num_rows - 1) * 16)
                preserve_relative_offset = True
            elif event.key() == Qt.Key_PageDown:
                new_cursor = min(self.end_addr - 1, self.cursor + (self.display_num_rows - 1) * 16)
                preserve_relative_offset = True
            elif event.key() == Qt.Key_Home:
                new_cursor = max(self.start_addr, self.cursor & ~0xF)
            elif event.key() == Qt.Key_End:
                new_cursor = min(self.end_addr - 1, (self.cursor & ~0xF) + 0xF)
            else:
                new_cursor = self.cursor + movement_keys[event.key()]
            if self.start_addr <= new_cursor < self.end_addr:
                self.move_viewport_to(new_cursor, preserve_relative_offset)
                self.set_cursor(new_cursor, update_viewport=False)
            event.accept()
            return
        elif QApplication.keyboardModifiers() & Qt.ControlModifier:
            if event.key() == Qt.Key_Space:
                self.set_cursor(self.cursor, ascii_column=not self.ascii_column_active)
                event.accept()
                return
        else:
            t = event.text()
            if len(t) == 1:
                if self.ascii_column_active:
                    if t.isascii() and t.isprintable():
                        self._set_byte_value(ord(t))
                        event.accept()
                        return
                else:
                    if t in "0123456789abcdefABCDEF":
                        self._set_nibble_value(int(t, 16))
                        event.accept()
                        return
        super().keyPressEvent(event)

    def _update_layout(self) -> None:
        """
        Update various layout settings based on font and data store
        """
        self.prepareGeometryChange()

        ti = QGraphicsSimpleTextItem()  # Get font metrics using text item
        ti.setFont(self.font)
        ti.setText("0")

        self.row_padding = int(ti.boundingRect().height() * 0.25)
        self.row_height = ti.boundingRect().height() + self.row_padding
        self.char_width = ti.boundingRect().width()
        self.section_space = self.char_width * 4
        self.addr_offset = self.char_width * 1
        self.addr_width = self.char_width * len(f"{self.display_end_addr:8x}")
        self.byte_width = self.char_width * 2
        self.byte_space = self.char_width * 1
        self.byte_group_space = self.char_width * 2
        self.ascii_width = self.char_width * 1
        self.ascii_space = 0

        self.byte_column_offsets = [self.addr_offset + self.addr_width + self.section_space]
        for i in range(1, 17):
            x = self.byte_column_offsets[-1] + self.byte_width + (self.byte_group_space if i == 8 else self.byte_space)
            self.byte_column_offsets.append(x)

        self.ascii_column_offsets = [self.byte_column_offsets[-1] + self.section_space]
        for _ in range(1, 17):
            x = self.ascii_column_offsets[-1] + self.ascii_width + self.ascii_space
            self.ascii_column_offsets.append(x)

        self.max_x = self.ascii_column_offsets[-1]
        self.max_y = self.display_num_rows * self.row_height

        self.update()

    def build_selection_path(
        self, min_addr: HexAddress, max_addr: HexAddress, ascii_section: bool = False, shrink: float = 0.0
    ) -> QPainterPath:
        """
        Build a QPainterPath that selects a given (inclusive addresses) range of bytes.
        """
        row_start, col_start = self.addr_to_row_col(min_addr)
        row_end, col_end = self.addr_to_row_col(max_addr)

        num_selected_rows = row_end - row_start + 1

        if ascii_section:
            column_offsets = self.ascii_column_offsets
            column_width = self.ascii_width
            column_space = self.ascii_space
        else:
            column_offsets = self.byte_column_offsets
            column_width = self.byte_width
            column_space = self.byte_space

        # Build top rect
        trect = QRectF()
        p = self.row_to_point(row_start)
        p.setX(column_offsets[col_start] - column_space / 2)
        trect.setTopLeft(p)
        p = QPointF(
            column_offsets[col_end if num_selected_rows == 1 else 15] + column_width + column_space / 2,
            p.y() + self.row_height,
        )
        trect.setBottomRight(p)
        trect = trect.marginsRemoved(QMarginsF(shrink, shrink, shrink, shrink))

        # Build middle rect
        if num_selected_rows > 2:
            mrect = QRectF()
            p = self.row_to_point(row_start + 1)
            p.setX(column_offsets[0] - column_space / 2)
            mrect.setTopLeft(p)
            p = QPointF(
                column_offsets[15] + column_width + column_space / 2, p.y() + self.row_height * (num_selected_rows - 2)
            )
            mrect.setBottomRight(p)
            mrect = mrect.marginsRemoved(QMarginsF(shrink, shrink, shrink, shrink))
        else:
            mrect = None

        # Build bottom rect
        if num_selected_rows > 1:
            brect = QRectF()
            p = self.row_to_point(row_end)
            p.setX(column_offsets[0] - column_space / 2)
            brect.setTopLeft(p)
            p = QPointF(column_offsets[col_end] + column_width + column_space / 2, p.y() + self.row_height)
            brect.setBottomRight(p)
            brect = brect.marginsRemoved(QMarginsF(shrink, shrink, shrink, shrink))
        else:
            brect = None

        close_to_top = True
        spath = QPainterPath()
        spath.moveTo(trect.topLeft())
        spath.lineTo(trect.topRight())
        spath.lineTo(trect.bottomRight())
        last = trect.bottomRight()
        if mrect:
            spath.lineTo(mrect.bottomRight())
            last = mrect.bottomRight()
        if brect:
            if mrect is None and col_end < col_start:
                # Don't try to connect disjoint top and bottom
                spath.lineTo(trect.bottomLeft())
                spath.closeSubpath()
                spath.moveTo(brect.topLeft())
                last = brect.topLeft()
                close_to_top = False
            spath.lineTo(QPointF(brect.topRight().x(), last.y()))
            spath.lineTo(brect.bottomRight())
            spath.lineTo(brect.bottomLeft())
            spath.lineTo(brect.topLeft())
            last = brect.topLeft()
        if mrect:
            spath.lineTo(mrect.topLeft())
            last = mrect.topLeft()
        if close_to_top:
            spath.lineTo(QPointF(trect.bottomLeft().x(), last.y()))
        spath.closeSubpath()
        return spath

    def get_selection(self) -> tuple[int, int] | None:
        """
        Get active selection, returning (minaddr, maxaddr) inclusive.
        """
        if self.selection_start is None:
            return None
        if self.start_addr <= self.cursor < self.end_addr:
            minaddr = min(self.cursor, self.selection_start)
            maxaddr = max(self.cursor, self.selection_start)
            return (minaddr, maxaddr)
        return None

    def toggle_cursor_blink(self) -> None:
        """
        Simply toggles cursor blink status.
        """
        self.cursor_blink_state = not self.cursor_blink_state
        self.update()

    def set_highlight_regions(self, regions: Sequence[HexHighlightRegion]) -> None:
        """
        Sets the list of highlighted regions.
        """
        self.highlighted_regions = regions
        self.update_active_highlight_regions()
        self.update()

    def paint_highlighted_region(self, painter, region: HexHighlightRegion) -> None:
        """
        Paint a highlighted region of bytes.
        """
        pen_width = 1.0
        half_pen_width = pen_width / 2.0
        end_addr = region.addr + region.size - 1

        path = self.build_selection_path(region.addr, end_addr, False, half_pen_width)

        color = QColor(region.color)
        if not region.active:
            color = color.darker(150)

        r = path.boundingRect()
        bg = PySide6.QtGui.QLinearGradient(r.topLeft(), r.bottomLeft())
        top_color = QColor(color)
        top_color.setAlpha(50)
        bg.setColorAt(0, top_color)
        bottom_color = QColor(color)
        bottom_color.setAlpha(10)
        bg.setColorAt(1, bottom_color)

        painter.setBrush(bg)
        painter.setPen(QPen(color, pen_width))
        painter.drawPath(path)
        painter.drawPath(self.build_selection_path(region.addr, end_addr, True, half_pen_width))

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        """
        Repaint the item.
        """
        min_row = self.point_to_row(option.exposedRect.topLeft())
        if min_row is None:
            min_row = 0
        max_row = self.point_to_row(option.exposedRect.bottomLeft())
        if max_row is None:
            max_row = self.display_num_rows - 1

        # Paint background
        painter.setPen(Qt.NoPen)
        for row in range(min_row, max_row + 1):
            row_addr = self.row_to_addr(row)
            if row_addr >= self.display_end_addr:
                break
            pt = self.row_to_point(row)
            painter.setBrush(Conf.palette_base if row_addr & 0x10 else Conf.palette_alternatebase)
            painter.drawRect(QRectF(0, pt.y(), self.boundingRect().width(), self.row_height))

        for region in self.highlighted_regions:
            self.paint_highlighted_region(painter, region)

        # Paint selection
        if self.selection_start is not None:
            minaddr = min(self.cursor, self.selection_start)
            maxaddr = max(self.cursor, self.selection_start)

            pen_width = 1.0
            half_pen_width = pen_width / 2.0

            def set_pen_brush_for_active_selection(active: bool) -> None:
                if active:
                    painter.setPen(QPen(Conf.hex_view_selection_color, pen_width))
                    painter.setBrush(QColor(255, 255, 255, 10))
                else:
                    painter.setPen(QPen(Conf.hex_view_selection_alt_color, pen_width))
                    painter.setBrush(QColor(255, 255, 255, 10))

            set_pen_brush_for_active_selection(not self.ascii_column_active)
            painter.drawPath(self.build_selection_path(minaddr, maxaddr, False, half_pen_width))

            set_pen_brush_for_active_selection(self.ascii_column_active)
            painter.drawPath(self.build_selection_path(minaddr, maxaddr, True, half_pen_width))

        # Paint text
        painter.setFont(self.font)

        for row in range(min_row, max_row + 1):
            row_addr = self.row_to_addr(row)
            if row_addr >= self.display_end_addr:
                break

            pt = self.row_to_point(row)
            pt.setY(pt.y() + self.row_height - self.row_padding)

            # Paint address
            addr_text = f"{row_addr:08x}"
            pt.setX(self.addr_offset)
            painter.setPen(Conf.disasm_view_node_address_color)
            painter.drawText(pt, addr_text)

            # Paint byte values
            for col in range(16):
                addr = self.row_col_to_addr(row, col)
                if addr < self.display_start_addr or addr >= self.display_end_addr:
                    continue
                val = self.read_func(addr)
                pt.setX(self.byte_column_offsets[col])

                if isinstance(val, int):
                    if is_printable(val):
                        color = Conf.disasm_view_printable_byte_color
                    else:
                        color = Conf.disasm_view_unprintable_byte_color
                    byte_text = f"{val:02x}"
                else:
                    byte_text = val * 2 if isinstance(val, str) and len(val) == 1 else "??"
                    color = Conf.disasm_view_unknown_byte_color

                pt.setX(self.byte_column_offsets[col])
                painter.setPen(color)
                painter.drawText(pt, byte_text)

            # Paint ASCII representation
            for col in range(16):
                addr = self.row_col_to_addr(row, col)
                if addr < self.display_start_addr or addr >= self.display_end_addr:
                    continue
                val = self.read_func(addr)
                pt.setX(self.ascii_column_offsets[col])

                if isinstance(val, int):
                    if is_printable(val):
                        color = Conf.disasm_view_printable_character_color
                        ch = chr(val)
                    else:
                        color = Conf.disasm_view_unprintable_character_color
                        ch = "."
                else:
                    color = Conf.disasm_view_unknown_character_color
                    ch = val if isinstance(val, str) and len(val) == 1 else "?"

                pt.setX(self.ascii_column_offsets[col])
                painter.setPen(color)
                painter.drawText(pt, ch)

        # Paint cursor
        if self.show_cursor and (self.display_start_addr <= self.cursor < self.display_end_addr):
            cursor_height = self.row_padding / 2

            def set_pen_brush_for_active_cursor(active: bool) -> None:
                painter.setPen(Qt.NoPen)
                if active:
                    col = Conf.palette_text if self.cursor_blink_state else Qt.NoBrush
                else:
                    col = Conf.palette_disabled_text
                painter.setBrush(col)

            # Byte cursor
            set_pen_brush_for_active_cursor(not self.ascii_column_active)
            tl = self.addr_to_point(self.cursor)
            tl.setY(tl.y() + self.row_height - cursor_height)
            cursor_width = self.byte_width
            if self.cursor_nibble is not None:
                cursor_width /= 2
                tl.setX(tl.x() + cursor_width * (1 - self.cursor_nibble))
            painter.drawRect(QRectF(tl, QSizeF(cursor_width, cursor_height)))

            # ASCII cursor
            set_pen_brush_for_active_cursor(self.ascii_column_active)
            tl = self.addr_to_point(self.cursor, True)
            tl.setY(tl.y() + self.row_height - cursor_height)
            painter.drawRect(QRectF(tl, QSizeF(self.ascii_width, cursor_height)))

    def boundingRect(self) -> PySide6.QtCore.QRectF:
        return QRectF(0, 0, self.max_x, self.max_y)

    def on_mouse_move_event_from_view(self, point: QPointF) -> None:
        """
        Highlight memory region under cursor.
        """
        self.setToolTip("")
        addr = self.point_to_addr(point)
        if addr is None:
            return
        addr, _ = addr
        regions = self.get_highlight_regions_at_addr(addr)
        if len(regions):
            t = "\n".join(t for t in [r.get_tooltip() for r in regions] if t)
            self.setToolTip(t)


class HexGraphicsSubView(QGraphicsView):
    """
    Wrapper QGraphicsView used for rendering and event propagation.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.setMouseTracking(True)
        self.setBackgroundBrush(Conf.palette_base)
        self.setResizeAnchor(QGraphicsView.NoAnchor)
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def wheelEvent(self, event) -> None:
        self.parent().wheelEvent(event)
        super().wheelEvent(event)

    def mouseMoveEvent(self, event) -> None:
        """
        Handle mouse move events.

        Mouse move events whilst not holding a mouse button will not be propagated down to QGraphicsItems, so we catch
        the movement events here in the view and forward them to the feature map item.
        """
        scene_pt = self.mapToScene(event.pos().x(), event.pos().y())
        item_pt = self.mapFromScene(scene_pt)
        self.parent().hex.on_mouse_move_event_from_view(item_pt)
        super().mouseMoveEvent(event)


class HexGraphicsView(QAbstractScrollArea):
    """
    Container view for the HexGraphicsObject.
    """

    cursor_changed = Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self._processing_scroll_event: bool = False

        self._view: QGraphicsView = HexGraphicsSubView(parent=self)
        self._scene: QGraphicsScene = QGraphicsScene(parent=self._view)
        self.hex: HexGraphicsObject = HexGraphicsObject()
        self.hex.cursor_changed.connect(self.on_cursor_changed)
        self._scene.addItem(self.hex)
        self._view.setScene(self._scene)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.verticalScrollBar().actionTriggered.connect(self._on_vertical_scroll_bar_triggered)
        self.horizontalScrollBar().actionTriggered.connect(self._on_horizontal_scroll_bar_triggered)
        self._view.setFrameStyle(QFrame.NoFrame)
        self.setFrameStyle(QFrame.NoFrame)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._view)
        self.setLayout(layout)

        self.scrollbar_range = 100
        self.verticalScrollBar().setRange(0, self.scrollbar_range)
        self._update_vertical_scrollbar()

    def _update_vertical_scrollbar(self) -> None:
        if self._processing_scroll_event:
            return
        addr_range = (self.hex.end_addr - self.hex.start_addr) >> 4
        if addr_range > 0:
            offset = (self.hex.display_start_addr - self.hex.start_addr) >> 4
            scrollbar_value = (offset + 1) / addr_range
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        else:
            scrollbar_value = 0
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.verticalScrollBar().setValue(int(scrollbar_value * self.scrollbar_range))

    def _update_horizontal_scrollbar(self) -> None:
        if self._processing_scroll_event:
            return
        hex_rect = self.hex.boundingRect()
        hex_rect.translate(self.hex.pos())
        vp_rect = self._view.sceneRect()
        scroll_range = max(0, int(hex_rect.width()) - int(vp_rect.width()))
        if scroll_range == 0:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        else:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            self.horizontalScrollBar().setRange(0, scroll_range)
            self.horizontalScrollBar().setPageStep(hex_rect.width() / 10)
            self.horizontalScrollBar().setValue(vp_rect.left())

    def _get_num_rows_visible(self, fully_visible: bool = False):
        num_rows_visible = int(self._view.mapToScene(0, self.viewport().height()).y() / self.hex.row_height)
        if not fully_visible:
            num_rows_visible += 1
        return num_rows_visible

    def update_display_num_rows(self) -> None:
        self.hex.set_display_num_rows(self._get_num_rows_visible())

    def set_display_offset(self, offset: HexAddress) -> None:
        self.hex.set_display_offset(offset)
        self._update_vertical_scrollbar()

    def get_display_start_addr(self) -> HexAddress:
        return self.hex.display_start_addr

    def set_display_start_addr(self, start: HexAddress) -> None:
        self.set_display_offset(start - self.hex.start_addr)

    def _on_vertical_scroll_bar_triggered(self, action: int) -> None:
        if self._processing_scroll_event:
            return
        self._processing_scroll_event = True
        action = QAbstractSlider.SliderAction(action)
        if action == QAbstractSlider.SliderSingleStepAdd:
            self.set_display_offset(self.hex.display_offset_addr + 0x10)
        elif action == QAbstractSlider.SliderSingleStepSub:
            self.set_display_offset(self.hex.display_offset_addr - 0x10)
        elif action == QAbstractSlider.SliderPageStepAdd:
            self.set_display_offset(self.hex.display_offset_addr + 0x10)
        elif action == QAbstractSlider.SliderPageStepSub:
            self.set_display_offset(self.hex.display_offset_addr - 0x10)
        elif action == QAbstractSlider.SliderMove:
            addr_range = self.hex.end_addr - self.hex.start_addr
            if addr_range <= 0:
                return

            sb_value = self.verticalScrollBar().value()
            if sb_value < 5:
                sb = 0
            elif sb_value > (self.scrollbar_range - 5):
                sb = 1.0
            else:
                sb = sb_value / self.scrollbar_range
            display_offset_addr = int(sb * addr_range) & ~0xF
            self.set_display_offset(display_offset_addr)
        self._processing_scroll_event = False

    def _on_horizontal_scroll_bar_triggered(self, action: int) -> None:
        self._processing_scroll_event = True
        action = QAbstractSlider.SliderAction(action)
        if action == QAbstractSlider.SliderMove:
            vp = self.viewport().geometry()
            vp.moveTo(0, 0)
            vp = self._view.mapToScene(vp).boundingRect()
            vp.moveTo(self.horizontalScrollBar().value(), 0)
            self._view.setSceneRect(vp)
        self._processing_scroll_event = False

    def wheelEvent(self, event: QWheelEvent) -> None:
        if event.modifiers() & Qt.ControlModifier == Qt.ControlModifier:
            self.adjust_viewport_scale(1.25 if event.angleDelta().y() > 0 else 1 / 1.25)
        else:
            d = event.angleDelta().y()
            if d != 0:
                self.set_display_offset(self.hex.display_offset_addr - 0x10 * d // 32)
        event.accept()

    def update_scene_rect(self) -> None:
        hex_rect = self.hex.boundingRect()
        hex_rect.translate(self.hex.pos())
        cursor_rect = self.hex.addr_to_rect(self.hex.cursor)
        cursor_rect.translate(self.hex.pos())
        vp_rect = self.viewport().geometry()
        vp_rect.moveTo(0, 0)
        vp_rect = self._view.mapToScene(vp_rect).boundingRect()

        if cursor_rect.right() > vp_rect.right():
            dX = vp_rect.right() - cursor_rect.right()  # Scroll right
        elif cursor_rect.left() < vp_rect.left():
            dX = vp_rect.left() - cursor_rect.left()  # Scroll left
        elif vp_rect.width() < hex_rect.width() and vp_rect.right() > hex_rect.right():
            dX = vp_rect.right() - hex_rect.right()  # Reveal left
        elif vp_rect.width() > hex_rect.width():
            dX = vp_rect.left()  # Reveal all
        else:
            dX = 0

        vp_rect.translate(-dX, 0)
        self._view.setSceneRect(vp_rect)
        self._update_horizontal_scrollbar()

    def resizeEvent(self, event: PySide6.QtGui.QResizeEvent) -> None:  # pylint: disable=unused-argument
        self._view.resize(self.viewport().size())
        self.update_scene_rect()
        self.update_display_num_rows()

    def set_region_callback(self, write, mem, addr: int, size: int) -> None:
        """
        Set current buffer.
        """
        self.hex.set_data_callback(write, mem, addr, size)
        self.update_scene_rect()
        self.set_display_offset(0)

    def clear(self) -> None:
        """
        Clear current buffer.
        """
        self.hex.clear()
        self.update_scene_rect()

    def on_cursor_changed(self) -> None:
        """
        Handle cursor change events.
        """
        self.cursor_changed.emit()
        self.update_scene_rect()
        self._update_vertical_scrollbar()

    def adjust_viewport_scale(self, scale: float | None = None) -> None:
        """
        Reset viewport scaling. If `scale` is None, viewport scaling is reset to default.
        """
        if scale is None:
            self._view.resetTransform()
        else:
            self._view.scale(scale, scale)
        self.update_scene_rect()
        self.update_display_num_rows()

    def changeEvent(self, event: QEvent) -> None:
        """
        Redraw on color scheme update.
        """
        if event.type() == QEvent.PaletteChange:
            self._view.setBackgroundBrush(Conf.palette_base)
            self.update()

    def keyPressEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        """
        Handle key events.
        """
        if event.modifiers() & Qt.ControlModifier == Qt.ControlModifier:
            if event.key() == Qt.Key_0:
                self.adjust_viewport_scale()
                event.accept()
                return
            elif event.key() == Qt.Key_Equal:
                self.adjust_viewport_scale(1.25)
                event.accept()
                return
            elif event.key() == Qt.Key_Minus:
                self.adjust_viewport_scale(1 / 1.25)
                event.accept()
                return
        super().keyPressEvent(event)


class HexView(SynchronizedInstanceView):
    """
    View and edit memory/object code in classic hex editor format.
    """

    _widgets_initialized: bool = False

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("hex", workspace, default_docking_position, instance)
        self.base_caption: str = "Hex"
        self.smart_highlighting_enabled: bool = True
        self._clipboard = None
        self._cfb_highlights: Sequence[HexHighlightRegion] = []
        self._sync_view_highlights: Sequence[HexHighlightRegion] = []
        self._patch_highlights: Sequence[PatchHighlightRegion] = []
        self._changed_data_highlights: Sequence[HexHighlightRegion] = []
        self._breakpoint_highlights: Sequence[BreakpointHighlightRegion] = []

        self._init_widgets()
        self.instance.cfb.am_subscribe(self._on_cfb_event)
        self.instance.patches.am_subscribe(self._update_highlight_regions_from_patches)
        self.instance.breakpoint_mgr.breakpoints.am_subscribe(self._update_highlight_regions_from_breakpoints)
        self._data_cache = {}

        self._reload_data()

        self._dbg_manager = self.instance.debugger_mgr
        self._dbg_watcher = DebuggerWatcher(self._on_debugger_state_updated, self._dbg_manager.debugger)
        self._on_debugger_state_updated()

    def _on_cfb_event(self, **kwargs) -> None:
        if not kwargs:
            self._reload_data()

    def closeEvent(self, event) -> None:
        self._dbg_watcher.shutdown()
        super().closeEvent(event)

    def _clear_highlights(self) -> None:
        """
        Clear all highlight regions
        """
        self._cfb_highlights = []
        self._sync_view_highlights = []
        self._patch_highlights = []
        self._changed_data_highlights = []
        self._breakpoint_highlights = []

    def _reload_data(self):
        """
        Callback when hex backing data store should be updated.
        """
        self._clear_highlights()
        start = self.inner_widget.get_display_start_addr()
        cursor = self.inner_widget.hex.cursor
        source = self._data_source_combo.currentData()
        if source == HexDataSource.Loader:
            if self.instance.cfb.am_none:
                self.inner_widget.clear()
                return
            loader = self.instance.project.loader
            self.inner_widget.set_region_callback(
                self.project_memory_write_func,
                self.project_memory_read_func,
                loader.min_addr,
                loader.max_addr - loader.min_addr + 1,
            )
            self._update_highlight_regions_from_patches()
        elif source == HexDataSource.Debugger:
            self._data_cache = {}
            self.inner_widget.set_region_callback(
                self.debugger_memory_write_func,
                self.debugger_memory_read_func,
                0,
                0x10000000000000000,  # FIXME: Get actual ranges and add them
            )
        else:
            raise NotImplementedError

        self.inner_widget.set_display_start_addr(start)
        self.inner_widget.hex.set_cursor(cursor)
        self._update_highlight_regions_from_synchronized_views()
        self._update_cfb_highlight_regions()
        self._set_highlighted_regions()

    def _data_source_changed(self, index: int) -> None:  # pylint:disable=unused-argument
        self._reload_data()

    def _on_debugger_state_updated(self) -> None:
        source = self._data_source_combo.currentData()
        if source == HexDataSource.Debugger:
            #
            # Calculate differences in state memory and highlight
            #
            previous_data = self._data_cache
            self._data_cache = {}
            addr = self.inner_widget.hex.display_start_addr
            regions = []
            r = None
            while addr < self.inner_widget.hex.display_end_addr:
                if addr in previous_data:
                    differs = self.debugger_memory_read_func(addr) != previous_data[addr]
                else:
                    differs = False
                if differs:
                    if r is None:
                        r = HexHighlightRegion(Qt.red, addr, 0)
                        regions.append(r)
                    r.size += 1
                else:
                    r = None
                addr += 1
            self._changed_data_highlights = regions
            self.inner_widget.hex.update()
            self._set_highlighted_regions()

    def debugger_memory_read_func(self, addr: int) -> HexByteValue:
        """
        Callback to populate hex view with bytes from debugger state.
        """
        if addr not in self._data_cache:
            dbg = self.instance.debugger_mgr.debugger
            if dbg.am_none:
                v = "?"
            else:
                state: angr.SimState | None = dbg.simstate
                if state is None:
                    v = "?"
                else:
                    try:
                        r = state.memory.load(addr, 1)
                        v = "S" if r.symbolic else state.solver.eval(r)
                    except Exception:  # pylint:disable=broad-except
                        log.exception("Failed to read @ %#x", addr)
                        v = "?"
            self._data_cache[addr] = v
        return self._data_cache[addr]

    def debugger_memory_write_func(self, addr: int, value: int) -> bool:  # pylint:disable=unused-argument,no-self-use
        """
        Callback to populate hex view with bytes.
        """
        # FIXME: For debuggers that support it, allow editing
        return False

    def project_memory_read_func(self, addr: int) -> HexByteValue:
        """
        Callback to populate hex view with bytes.
        """
        p = self.instance.project

        patches = p.kb.patches.get_all_patches(addr, 1)
        if len(patches) > 0:
            patch = patches[0]
            return patch.new_bytes[addr - patch.addr]

        try:
            return p.loader.memory[addr]
        except KeyError:
            return "?"

    def auto_patch(self, addr: int, new_bytes: bytearray):
        """
        Automatically update or create patches to ensure new_bytes are patched at addr.
        """
        pm = self.instance.project.kb.patches
        max_addr = addr + len(new_bytes) - 1

        for p in pm.get_all_patches(addr, len(new_bytes)):
            patch_max_addr = p.addr + len(p) - 1
            if (p.addr <= addr) and (patch_max_addr >= max_addr):
                # Existing patch contains new patch entirely. Update it.
                p.new_bytes = p.new_bytes[: addr - p.addr] + new_bytes + p.new_bytes[max_addr - p.addr + 1 :]
                return
            elif (p.addr >= addr) and (patch_max_addr <= max_addr):
                # Patch will be entirely overwritten, remove it.
                pm.remove_patch(p.addr)
            elif (p.addr >= addr) and (patch_max_addr > max_addr):
                # Lower portion of patch will be overwritten, shrink patch up.
                pm.remove_patch(p.addr)
                new_p_addr = max_addr + 1
                p.new_bytes = p.new_bytes[(new_p_addr - p.addr) :]
                p.addr = new_p_addr
                pm.add_patch_obj(p)
            elif (p.addr < addr) and (patch_max_addr <= max_addr):
                # Upper portion of patch will be overwritten, shrink patch down.
                pm.remove_patch(p.addr)
                p.new_bytes = p.new_bytes[0 : (addr - p.addr)]
                pm.add_patch_obj(p)
            else:
                raise AssertionError

        # Check to see if we should extend an adjacent patch
        if addr > 0:
            p = pm.get_all_patches(addr - 1, 1)
            if len(p) > 0:
                p = p[0]
                p.new_bytes += new_bytes
                return

        pm.add_patch_obj(Patch(addr, new_bytes))

    def project_memory_write_bytearray(self, addr: int, value: bytearray) -> bool:
        """
        Callback to write array of bytes as patch.
        """
        self.auto_patch(addr, bytearray(value))
        pm = self.instance.project.kb.patches
        pm_notifier = self.instance.patches
        if pm_notifier.am_none:
            pm_notifier.am_obj = pm
        pm_notifier.am_event()
        return True

    def project_memory_write_func(self, addr: int, value: int) -> bool:
        """
        Callback to populate hex view with bytes.
        """
        return self.project_memory_write_bytearray(addr, bytearray([value]))

    def set_smart_highlighting_enabled(self, enable: bool) -> None:
        """
        Control whether smart highlighting is enabled or not.
        """
        self.smart_highlighting_enabled = enable
        self._update_cfb_highlight_regions()

    def _init_widgets(self) -> None:
        """
        Initialize widgets for this view.
        """
        window = QMainWindow()
        window.setWindowFlags(Qt.Widget)

        status_bar = QFrame()
        status_lyt = QHBoxLayout()
        status_lyt.setContentsMargins(3, 3, 3, 3)
        status_lyt.setSpacing(3)

        self._status_lbl = QLabel()
        self._status_lbl.setText("Address: ")

        status_lyt.addWidget(self._status_lbl)
        status_lyt.addStretch(0)

        self._data_source_combo = QComboBox(self)
        self._data_source_combo.addItem("Loader", HexDataSource.Loader)
        self._data_source_combo.addItem("Debugger", HexDataSource.Debugger)
        self._data_source_combo.activated.connect(self._data_source_changed)
        status_lyt.addWidget(self._data_source_combo)

        option_btn = QPushButton()
        option_btn.setText("Options")
        option_mnu = QMenu(self)
        smart_hl_act = QAction("Smart &highlighting", self)
        smart_hl_act.setCheckable(True)
        smart_hl_act.setChecked(self.smart_highlighting_enabled)
        smart_hl_act.toggled.connect(self.set_smart_highlighting_enabled)
        option_mnu.addAction(smart_hl_act)
        option_btn.setMenu(option_mnu)
        status_lyt.addWidget(option_btn)

        status_bar.setLayout(status_lyt)

        self.inner_widget = HexGraphicsView(parent=self)
        lyt = QVBoxLayout()
        lyt.addWidget(status_bar)
        lyt.addWidget(self.inner_widget)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.setSpacing(0)
        self.setLayout(lyt)
        self.inner_widget.cursor_changed.connect(self.on_cursor_changed)
        self.inner_widget.hex.viewport_changed.connect(self.on_cursor_changed)

        self._widgets_initialized = True

    def revert_selected_patches(self) -> None:
        """
        Revert any selected patches.
        """
        dlg = QMessageBox()
        dlg.setWindowTitle("Revert patches")
        dlg.setText("Are you sure you want to revert selected patches?")
        dlg.setIcon(QMessageBox.Question)
        dlg.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
        dlg.setDefaultButton(QMessageBox.Cancel)
        if dlg.exec_() != QMessageBox.Yes:
            return

        selected_regions = self.inner_widget.hex.get_active_highlight_regions()
        for r in selected_regions:
            if isinstance(r, PatchHighlightRegion):
                r.revert()

    def _can_merge_any_selected_patches(self):
        """
        Determine if any of the selected patches can be merged.
        """
        return self._merge_selected_patches(True)

    def _merge_selected_patches(self, trial_only: bool = False) -> bool:
        """
        Merge selected directly-adjacent patches.
        """
        selected_patches = [
            r for r in self.inner_widget.hex.get_active_highlight_regions() if isinstance(r, PatchHighlightRegion)
        ]
        i = 0
        did_patch = False
        while i < (len(selected_patches) - 1):
            patch = selected_patches[i]
            for j, neighbor in enumerate(selected_patches[i + 1 :]):
                if patch.can_merge_with(neighbor):
                    if trial_only:
                        return True
                    patch.merge_with(neighbor)
                    did_patch = True
                else:
                    i = i + j + 1
                    break

        return did_patch

    def _get_num_selected_bytes(self) -> int:
        """
        Determine whether any bytes are selected.
        """
        sel = self.inner_widget.hex.get_selection()
        if sel is None:
            return 0
        minaddr, maxaddr = sel
        num_bytes_selected = maxaddr - minaddr + 1
        return num_bytes_selected

    def _copy_selected_bytes(self) -> None:
        """
        Copy selected bytes to view-only clipboard.
        """
        sel = self.inner_widget.hex.get_selection()
        if sel is None:
            self._clipboard = None
            return

        minaddr, maxaddr = sel
        num_bytes_selected = maxaddr - minaddr + 1

        self._clipboard = bytearray(num_bytes_selected)
        for addr in range(minaddr, maxaddr + 1):
            d = self.project_memory_read_func(addr)  # FIXME: Support multibyte read
            if isinstance(d, int):
                self._clipboard[addr - minaddr] = d

    def _paste_copied_bytes_at_cursor(self) -> None:
        """
        Paste the view-only clipboard contents at cursor location.
        """
        if self._clipboard is None:
            return
        if self._data_source_combo.currentData() == HexDataSource.Loader:
            self.project_memory_write_bytearray(self.inner_widget.hex.cursor, self._clipboard)
        # FIXME: Support pasting data to current debugger state

    def _set_breakpoint(self, bp_type: BreakpointType = BreakpointType.Execute) -> None:
        """
        Set breakpoint at current cursor.
        """
        sel = self.inner_widget.hex.get_selection()
        if sel:
            minaddr, maxaddr = sel
            num_bytes_selected = maxaddr - minaddr + 1
        else:
            minaddr = self.inner_widget.hex.cursor
            num_bytes_selected = 1
        self.instance.breakpoint_mgr.add_breakpoint(Breakpoint(bp_type, minaddr, num_bytes_selected))

    def _get_breakpoint_submenu(self) -> QMenu:
        """
        Get context menu to add new breakpoints.
        """
        mnu = QMenu("Set &breakpoint", self)
        act = QAction("Break on &Execute", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Execute))
        mnu.addAction(act)
        act = QAction("Break on &Read", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Read))
        mnu.addAction(act)
        act = QAction("Break on &Write", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Write))
        mnu.addAction(act)
        return mnu

    def contextMenuEvent(self, event: PySide6.QtGui.QContextMenuEvent) -> None:  # pylint: disable=unused-argument
        """
        Display view context menu.
        """
        mnu = QMenu(self)
        add_sep = False

        # FIXME: This should also go into an Edit menu accessible from the main window
        num_selected_bytes = self._get_num_selected_bytes()
        if num_selected_bytes > 0:
            plural = "s" if num_selected_bytes != 1 else ""
            act = QAction(f"Copy {num_selected_bytes:d} byte{plural}", mnu)
            act.triggered.connect(self._copy_selected_bytes)
            mnu.addAction(act)
            add_sep = True
        if self._clipboard is not None and self._data_source_combo.currentData() == HexDataSource.Loader:
            plural = "s" if len(self._clipboard) != 1 else ""
            act = QAction(f"Paste {len(self._clipboard):d} byte{plural}", mnu)
            act.triggered.connect(self._paste_copied_bytes_at_cursor)
            mnu.addAction(act)
            add_sep = True

        if add_sep:
            mnu.addSeparator()
            add_sep = False

        mnu.addMenu(self._get_breakpoint_submenu())
        mnu.addSeparator()

        # Get context menu for specific item under cursor
        for rgn in self.inner_widget.hex.get_highlight_regions_under_cursor():
            rgn_mnu = rgn.gen_context_menu_actions()
            if rgn_mnu is not None:
                mnu.addMenu(rgn_mnu)
                add_sep = True

        if add_sep:
            mnu.addSeparator()
            add_sep = False

        # Get context menu for groups of items
        selected_regions = self.inner_widget.hex.get_active_highlight_regions()
        if any(isinstance(r, PatchHighlightRegion) for r in selected_regions):
            act = QAction("Merge selected patches", mnu)
            act.triggered.connect(self._merge_selected_patches)
            act.setEnabled(self._can_merge_any_selected_patches())
            mnu.addAction(act)
            act = QAction("Revert selected patches", mnu)
            act.triggered.connect(self.revert_selected_patches)
            mnu.addAction(act)
            add_sep = True

        if add_sep:
            mnu.addSeparator()

        mnu.addMenu(self.get_synchronize_with_submenu())
        mnu.exec_(QCursor.pos())

    def set_cursor(self, addr: int) -> None:
        """
        Move cursor to specific address and clear any active selection.
        """
        self.inner_widget.hex.clear_selection()
        self.inner_widget.hex.set_cursor(addr)

    def on_cursor_changed(self) -> None:
        """
        Handle updates to cursor or viewport.
        """
        self.update_status_text()
        self._update_cfb_highlight_regions()
        self.set_synchronized_cursor_address(self.inner_widget.hex.cursor)
        self.published_view_state.cursors = [self.inner_widget.hex.cursor]
        self.notify_view_state_updated()

    def update_status_text(self) -> None:
        """
        Update status text with current cursor info.
        """
        sel = self.inner_widget.hex.get_selection()
        if sel:
            minaddr, maxaddr = sel
            bytes_selected = maxaddr - minaddr + 1
            plural = "s" if bytes_selected != 1 else ""
            s = f"Address: [{minaddr:08x}, {maxaddr:08x}], {bytes_selected} byte{plural} selected"
        else:
            s = f"Address: {self.inner_widget.hex.cursor:08x}"
        self._status_lbl.setText(s)

    def keyPressEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        """
        Handle key events.
        """
        if event.key() == Qt.Key_G:
            self.popup_jumpto_dialog()
            return

        super().keyPressEvent(event)

    def popup_jumpto_dialog(self) -> None:
        """
        Display 'Jump To' dialog.
        """
        JumpTo(self, parent=self).exec_()

    def jump_to(self, addr: int) -> bool:
        """
        Jump to a specific address.
        """
        self.set_cursor(addr)
        return True

    def on_synchronized_view_group_changed(self) -> None:
        """
        Handle view being added to or removed from the view synchronization group.
        """
        if self._widgets_initialized:
            self.inner_widget.hex.set_always_show_cursor(len(self.sync_state.views) > 1)

    def on_synchronized_highlight_regions_changed(self) -> None:
        """
        Handle synchronized highlight region change event.
        """
        self._update_highlight_regions_from_synchronized_views()

    def _update_cfb_highlight_regions(self) -> None:
        """
        Update cached list of highlight regions under cursor.
        """
        regions = []
        cfb = self.instance.cfb
        if self.smart_highlighting_enabled and not cfb.am_none:
            for item in cfb.floor_items(self.inner_widget.hex.display_start_addr):
                item_addr, item = item
                if item.size is None:
                    continue
                if (item_addr + item.size) < self.inner_widget.hex.display_start_addr:
                    continue
                if item_addr >= self.inner_widget.hex.display_end_addr:
                    break
                if isinstance(item, MemoryData):
                    is_string = item.sort in (MemoryDataSort.String, MemoryDataSort.UnicodeString)
                    color = Conf.hex_view_string_color if is_string else Conf.hex_view_data_color
                    regions.append(HexHighlightRegion(color, item.addr, item.size, str(item)))
                elif isinstance(item, Block):
                    try:
                        for insn in item.disassembly.insns:
                            s = f"{insn} in {item}"
                            regions.append(
                                HexHighlightRegion(Conf.hex_view_instruction_color, insn.address, insn.size, s)
                            )
                    except angr.errors.SimEngineError:
                        pass  # We may get a node in CFB to a non-decodeable address
        self._cfb_highlights = regions
        self._set_highlighted_regions()

    def _update_highlight_regions_from_synchronized_views(self) -> None:
        """
        Update cached list of highlight regions from synchronized views.
        """
        regions = []
        for v in self.sync_state.highlight_regions:
            if v is not self:
                for r in self.sync_state.highlight_regions[v]:
                    regions.append(HexHighlightRegion(Qt.green, r.addr, r.size))

        self._sync_view_highlights = regions
        self._set_highlighted_regions()

    def _update_highlight_regions_from_patches(self, **_) -> None:
        """
        Updates cached list of highlight regions from patches.
        """
        if self.instance.project.am_none:
            self._patch_highlights = []
        else:
            self._patch_highlights = [
                PatchHighlightRegion(patch, self) for patch in self.instance.project.kb.patches.values()
            ]
        self._set_highlighted_regions()

    def _update_highlight_regions_from_breakpoints(self, **kwargs) -> None:  # pylint:disable=unused-argument
        """
        Updates cached list of highlight regions from breakpoints.
        """
        if self.instance.project.am_none:
            self._breakpoint_highlights = []
        else:
            self._breakpoint_highlights = [
                BreakpointHighlightRegion(bp, self) for bp in self.instance.breakpoint_mgr.breakpoints
            ]
        self._set_highlighted_regions()

    def _set_highlighted_regions(self) -> None:
        """
        Update highlighted regions, with data from CFB and synchronized views.
        """
        regions = []
        regions.extend(self._cfb_highlights)
        regions.extend(self._sync_view_highlights)
        source = self._data_source_combo.currentData()
        if source == HexDataSource.Loader:
            regions.extend(self._patch_highlights)
        elif source == HexDataSource.Debugger:
            regions.extend(self._changed_data_highlights)
        regions.extend(self._breakpoint_highlights)
        self.inner_widget.hex.set_highlight_regions(regions)



from __future__ import annotations

import contextlib
import logging
import os
import re
from typing import TYPE_CHECKING, Any

import tomlkit
import tomlkit.exceptions
import tomlkit.items
from PySide6.QtGui import QColor, QFont, QFontMetricsF
from PySide6.QtWidgets import QApplication, QMessageBox

from angrmanagement.utils.env import app_root

from .config_entry import ConfigurationEntry as CE

if TYPE_CHECKING:
    import enum
    from collections.abc import Callable

_l = logging.getLogger(__name__)
color_re = re.compile("[0-9a-fA-F]+")


class UninterpretedCE(CE):
    """
    A config entry which has not been parsed because no type was available for it.
    """

    def __init__(self, name: str, value, default_value=None) -> None:
        super().__init__(name, UninterpretedCE, value, default_value=default_value)


def tomltype2pytype(v, ty: type | None) -> Any:
    if ty is str:
        if not isinstance(v, tomlkit.items.String):
            raise TypeError
        return str(v)
    elif ty is int:
        if not isinstance(v, tomlkit.items.Integer):
            raise TypeError
        return v.unwrap()
    elif ty is list:
        if not isinstance(v, tomlkit.items.Array):
            raise TypeError
        return [tomltype2pytype(v_, None) for v_ in v.value]
    return str(v) if isinstance(v, tomlkit.items.String) else v.unwrap()


def color_parser(config_option, value) -> QColor | None:
    if not isinstance(value, str) or not color_re.match(value) or len(value) not in (3, 6, 8, 12):
        _l.error("Failed to parse value %r as rgb color for option %s", value, config_option)
        return None

    return QColor("#" + value)


def color_serializer(config_option, value: QColor) -> str:
    if not isinstance(value, QColor):
        _l.error("Failed to serialize value %r as rgb color for option %s", value, config_option)
        return None

    return f"{value.alpha():02x}{value.red():02x}{value.green():02x}{value.blue():02x}"


def font_parser(config_option, value) -> QFont | None:
    if not isinstance(value, str) or "px " not in value:
        _l.error("Failed to parse value %r as font for option %s", value, config_option)
        return None

    parts = value.split("px ", 1)
    try:
        size = int(parts[0])
    except ValueError:
        _l.error("Failed to parse value %r as font for option %s", value, config_option)
        return None

    return QFont(parts[1], size)


def font_serializer(config_option, value: QFont) -> str:
    if not isinstance(value, QFont):
        _l.error("Failed to serialize value %r as font for option %s", value, config_option)
        return None

    return f"{value.pointSize()}px {value.family()}"


def enum_parser_serializer_generator(
    the_enum: enum.Enum, default
) -> tuple[Callable[[str, str], enum.Enum], Callable[[str, enum.Enum], str]]:
    def parser(config_option: str, value: str) -> enum.Enum:
        try:
            return the_enum[value]
        except KeyError:
            _l.error(
                "Failed to parse value %r as %s for option %s. Default to %s.",
                value,
                type(the_enum),
                config_option,
                default,
            )
        return default

    def serializer(config_option: str, value: enum.Enum) -> str:
        if not isinstance(value, the_enum):
            _l.error(
                "Failed to serialize value %r as %s for option %s. Default to %s.",
                value,
                type(the_enum),
                config_option,
                default,
            )
            return default
        return value.name

    return parser, serializer


def bool_parser(config_option, value) -> bool:
    if not value:
        return False
    if value.lower() in {"true", "1", "yes"}:
        return True
    if value.lower() in {"false", "0", "no"}:
        return False
    _l.error("Failed to parse value %r as bool for option %s. Default to False.", value, config_option)
    return False


def bool_serializer(config_option, value: bool) -> str:
    if not isinstance(value, bool):
        _l.error("Failed to serialize value %r as bool for option %s. Default to False.", value, config_option)
        return "false"
    return "true" if value else "false"


data_serializers = {
    QColor: (color_parser, color_serializer),
    QFont: (font_parser, font_serializer),
    bool: (bool_parser, bool_serializer),
    QFont.Weight: enum_parser_serializer_generator(QFont.Weight, QFont.Weight.Normal),
    QFont.Style: enum_parser_serializer_generator(QFont.Style, QFont.Style.StyleNormal),
}


# CE(name, type, default_value)
ENTRIES = [
    CE("ui_default_font", QFont, None),
    CE("tabular_view_font", QFont, None),
    CE("disasm_font", QFont, QFont("DejaVu Sans Mono", 10)),
    CE("symexec_font", QFont, QFont("DejaVu Sans Mono", 10)),
    CE("code_font", QFont, QFont("Source Code Pro", 10)),
    CE("theme_name", str, "Light"),
    CE("disasm_view_minimap_viewport_color", QColor, QColor(0xFF, 0x00, 0x00)),
    CE("disasm_view_minimap_background_color", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("disasm_view_minimap_outline_color", QColor, QColor(0xB8, 0xB8, 0xB8, 0xFF)),
    CE("disasm_view_background_color", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("disasm_view_operand_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("disasm_view_operand_constant_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("disasm_view_variable_label_color", QColor, QColor(0x00, 0x80, 0x00)),
    CE("disasm_view_operand_highlight_color", QColor, QColor(0xFC, 0xEF, 0x00)),
    CE("disasm_view_operand_select_color", QColor, QColor(0xFF, 0xFF, 0x00)),
    CE("disasm_view_function_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("disasm_view_string_color", QColor, QColor(0xA0, 0xA0, 0xA4)),
    CE("disasm_view_variable_ident_color", QColor, QColor(0xAA, 0x25, 0xDA)),
    CE("disasm_view_variable_offset_color", QColor, QColor(0x80, 0x80, 0x00)),
    CE("disasm_view_branch_target_text_color", QColor, QColor(0x80, 0x80, 0x00)),
    CE("disasm_view_comment_color", QColor, QColor(0x37, 0x3D, 0x3F, 0xFF)),
    CE("disasm_view_ir_default_color", QColor, QColor(0x80, 0x80, 0x80)),
    CE("disasm_view_label_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("disasm_view_label_highlight_color", QColor, QColor(0xF0, 0xF0, 0xBF)),
    CE("disasm_view_target_addr_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("disasm_view_antitarget_addr_color", QColor, QColor(0xFF, 0x00, 0x00)),
    CE("disasm_view_node_shadow_color", QColor, QColor(0x00, 0x00, 0x00, 0x00)),
    CE("disasm_view_node_background_color", QColor, QColor(0xFA, 0xFA, 0xFA)),
    CE("disasm_view_node_zoomed_out_background_color", QColor, QColor(0xDA, 0xDA, 0xDA)),
    CE("disasm_view_node_border_color", QColor, QColor(0xF0, 0xF0, 0xF0)),
    CE("disasm_view_node_instruction_selected_background_color", QColor, QColor(0xB8, 0xC3, 0xD6)),
    CE("disasm_view_node_address_color", QColor, QColor(0x00, 0x00, 0x00)),
    CE("disasm_view_node_mnemonic_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("disasm_view_node_rounding", int, 0),
    CE("disasm_view_selected_node_border_color", QColor, QColor(0x6B, 0x71, 0x7C)),
    CE("disasm_view_printable_byte_color", QColor, QColor(0x00, 0x80, 0x40)),
    CE("disasm_view_printable_character_color", QColor, QColor(0x00, 0x80, 0x40)),
    CE("disasm_view_unprintable_byte_color", QColor, QColor(0x80, 0x40, 0x00)),
    CE("disasm_view_unprintable_character_color", QColor, QColor(0x80, 0x40, 0x00)),
    CE("disasm_view_unknown_byte_color", QColor, QColor(0xF0, 0x00, 0x00)),
    CE("disasm_view_unknown_character_color", QColor, QColor(0xF0, 0x00, 0x00)),
    CE("disasm_view_back_edge_color", QColor, QColor(0xF9, 0xD5, 0x77)),
    CE("disasm_view_true_edge_color", QColor, QColor(0x79, 0xCC, 0xCD)),
    CE("disasm_view_false_edge_color", QColor, QColor(0xF1, 0x66, 0x64)),
    CE("disasm_view_direct_jump_edge_color", QColor, QColor(0x56, 0x5A, 0x5C)),
    CE("disasm_view_exception_edge_color", QColor, QColor(0xF9, 0x91, 0x0A)),
    CE("hex_view_selection_color", QColor, QColor(0xFF, 0x00, 0x00)),
    CE("hex_view_selection_alt_color", QColor, QColor(0xA0, 0xA0, 0xA4)),
    CE("hex_view_data_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("hex_view_string_color", QColor, QColor(0x00, 0xFF, 0xFF)),
    CE("hex_view_instruction_color", QColor, QColor(0xFF, 0x00, 0xFF)),
    CE("function_table_color", QColor, QColor(0x00, 0x00, 0x00)),
    CE("function_table_syscall_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("function_table_plt_color", QColor, QColor(0x00, 0x80, 0x00)),
    CE("function_table_simprocedure_color", QColor, QColor(0x80, 0x00, 0x00)),
    CE("function_table_alignment_color", QColor, QColor(0x80, 0x00, 0x80)),
    CE("function_table_signature_bg_color", QColor, QColor(0xAA, 0xFF, 0xFF)),
    CE("palette_window", QColor, QColor(0xEF, 0xEF, 0xEF, 0xFF)),
    CE("palette_windowtext", QColor, QColor(0x00, 0x00, 0x00, 0xFF)),
    CE("palette_base", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("palette_alternatebase", QColor, QColor(0xF7, 0xF7, 0xF7, 0xFF)),
    CE("palette_tooltipbase", QColor, QColor(0xFF, 0xFF, 0xDC, 0xFF)),
    CE("palette_tooltiptext", QColor, QColor(0x00, 0x00, 0x00, 0xFF)),
    CE("palette_placeholdertext", QColor, QColor(0x00, 0x00, 0x00, 0xAF)),
    CE("palette_text", QColor, QColor(0x00, 0x00, 0x00, 0xFF)),
    CE("palette_button", QColor, QColor(0xEF, 0xEF, 0xEF, 0xFF)),
    CE("palette_buttontext", QColor, QColor(0x00, 0x00, 0x00, 0xFF)),
    CE("palette_brighttext", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("palette_highlight", QColor, QColor(0x30, 0x8C, 0xC6, 0xFF)),
    CE("palette_highlightedtext", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("palette_disabled_text", QColor, QColor(0xBE, 0xBE, 0xBE, 0xFF)),
    CE("palette_disabled_buttontext", QColor, QColor(0xBE, 0xBE, 0xBE, 0xFF)),
    CE("palette_disabled_windowtext", QColor, QColor(0xBE, 0xBE, 0xBE, 0xFF)),
    CE("palette_light", QColor, QColor(0xFF, 0xFF, 0xFF, 0xFF)),
    CE("palette_midlight", QColor, QColor(0xCA, 0xCA, 0xCA, 0xFF)),
    CE("palette_dark", QColor, QColor(0x9F, 0x9F, 0x9F, 0xFF)),
    CE("palette_mid", QColor, QColor(0xB8, 0xB8, 0xB8, 0xFF)),
    CE("palette_shadow", QColor, QColor(0x76, 0x76, 0x76, 0xFF)),
    CE("palette_link", QColor, QColor(0x00, 0x00, 0xFF, 0xFF)),
    CE("palette_linkvisited", QColor, QColor(0xFF, 0x00, 0xFF, 0xFF)),
    CE("pseudocode_comment_color", QColor, QColor(0x00, 0x80, 0x00, 0xFF)),
    CE("pseudocode_comment_weight", QFont.Weight, QFont.Weight.Bold),
    CE("pseudocode_comment_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_function_color", QColor, QColor(0x00, 0x00, 0xFF, 0xFF)),
    CE("pseudocode_library_function_color", QColor, QColor(0xFF, 0x00, 0xFF)),
    CE("pseudocode_function_weight", QFont.Weight, QFont.Weight.Bold),
    CE("pseudocode_function_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_quotation_color", QColor, QColor(0x00, 0x80, 0x00, 0xFF)),
    CE("pseudocode_quotation_weight", QFont.Weight, QFont.Weight.Normal),
    CE("pseudocode_quotation_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_keyword_color", QColor, QColor(0x00, 0x00, 0x80, 0xFF)),
    CE("pseudocode_keyword_weight", QFont.Weight, QFont.Weight.Bold),
    CE("pseudocode_keyword_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_types_color", QColor, QColor(0x00, 0x00, 0x80, 0xFF)),
    CE("pseudocode_types_weight", QFont.Weight, QFont.Weight.Normal),
    CE("pseudocode_types_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_variable_color", QColor, QColor(0x00, 0x00, 0x00, 0xFF)),
    CE("pseudocode_variable_weight", QFont.Weight, QFont.Weight.Normal),
    CE("pseudocode_variable_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_label_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("pseudocode_label_weight", QFont.Weight, QFont.Weight.Normal),
    CE("pseudocode_label_style", QFont.Style, QFont.Style.StyleNormal),
    CE("pseudocode_highlight_color", QColor, QColor(0xFF, 0xFF, 0x00, 0xFF)),
    CE("proximity_node_background_color", QColor, QColor(0xFA, 0xFA, 0xFA)),
    CE("proximity_node_selected_background_color", QColor, QColor(0xCC, 0xCC, 0xCC)),
    CE("proximity_node_border_color", QColor, QColor(0xF0, 0xF0, 0xF0)),
    CE("proximity_function_node_text_color", QColor, QColor(0xFF, 0x00, 0x00)),
    CE("proximity_string_node_text_color", QColor, QColor(0x00, 0x80, 0x00)),
    CE("proximity_integer_node_text_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("proximity_variable_node_text_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("proximity_unknown_node_text_color", QColor, QColor(0x00, 0x00, 0x80)),
    CE("proximity_call_node_text_color", QColor, QColor(0x00, 0x00, 0xFF)),
    CE("proximity_call_node_text_color_plt", QColor, QColor(0x8B, 0x00, 0x8B)),
    CE("proximity_call_node_text_color_simproc", QColor, QColor(0x8B, 0x00, 0x8B)),
    CE("log_timestamp_format", str, "%X"),
    # FLIRT signatures
    CE("flirt_signatures_root", str, "./flirt_signatures/"),
    # Library documentation
    CE("library_docs_root", str, "./library_docs/"),
    # feature map
    CE("feature_map_regular_function_color", QColor, QColor(0x00, 0xA0, 0xE8)),
    CE("feature_map_unknown_color", QColor, QColor(0x0A, 0x0A, 0x0A)),
    CE("feature_map_delimiter_color", QColor, QColor(0x00, 0x00, 0x00)),
    CE("feature_map_data_color", QColor, QColor(0xC0, 0xC0, 0xC0)),
    CE("feature_map_string_color", QColor, QColor(0x00, 0xF0, 0x80)),
    # networking
    CE("http_proxy", str, ""),
    CE("https_proxy", str, ""),
    # plugins
    CE("plugin_search_path", str, "$AM_BUILTIN_PLUGINS:~/.local/share/angr-management/plugins"),
    CE("enabled_plugins", str, ""),
    # configurations for individual plugins
    # TOOD: Move them to separate locations
    CE("plugin_ComponentsPlugin_enabled", bool, False),
    # VARec
    CE("varec_endpoint", str, "http://192.168.32.129:5000/varec_joint_small"),
    # Daemon
    CE("use_daemon", bool, False),
    # Tabs
    CE("enabled_tabs", str, ""),
    # Recent
    CE("recent_files", list, []),
    CE("prompted_for_url_scheme_registration", bool, False),
]


class ConfigurationManager:  # pylint: disable=assigning-non-slot
    """
    Globe Configuration Manager for UI configuration with save/load function
    """

    __slots__ = (
        "_entries",
        "_disasm_font",
        "_disasm_font_metrics",
        "_disasm_font_height",
        "_disasm_font_width",
        "_disasm_font_ascent",
        "_symexec_font",
        "_symexec_font_metrics",
        "_symexec_font_height",
        "_symexec_font_width",
        "_symexec_font_ascent",
        "_code_font",
        "_code_font_metrics",
        "_code_font_height",
        "_code_font_width",
        "_code_font_ascent",
    )

    def __init__(self, entries=None) -> None:
        self._disasm_font = self._disasm_font_metrics = self._disasm_font_height = None
        self._disasm_font_width = self._disasm_font_ascent = None
        self._symexec_font = self._symexec_font_metrics = self._symexec_font_height = None
        self._symexec_font_width = self._symexec_font_ascent = None
        self._code_font = self._code_font_metrics = self._code_font_height = None
        self._code_font_width = self._code_font_ascent = None

        if entries is None:
            self._entries = {}
            self.load_initial_entries(reset=True)
        else:
            self._entries = entries

    def load_initial_entries(self, reset: bool = True) -> None:
        """
        Load configuration entries into self._entries.

        :param reset:   Reset all configuration items to their default values.
        :return:        None
        """
        for entry in ENTRIES:
            if entry.name not in self._entries:
                self._entries[entry.name] = entry.copy()
            else:
                if reset:
                    self._entries[entry.name] = entry.copy()

    @staticmethod
    def _manage_font_cache(real_font, font, metrics, height, width: int, ascent):
        if real_font == font:
            return font, metrics, height, width, ascent

        metrics = QFontMetricsF(real_font)
        height = metrics.height()
        width = metrics.width("A")
        ascent = metrics.ascent()
        return real_font, metrics, height, width, ascent

    def _disasm_manage_font_cache(self) -> None:
        (
            self._disasm_font,
            self._disasm_font_metrics,
            self._disasm_font_height,
            self._disasm_font_width,
            self._disasm_font_ascent,
        ) = ConfigurationManager._manage_font_cache(
            self.disasm_font,
            self._disasm_font,
            self._disasm_font_metrics,
            self._disasm_font_height,
            self._disasm_font_width,
            self._disasm_font_ascent,
        )

    def _symexec_manage_font_cache(self) -> None:
        (
            self._symexec_font,
            self._symexec_font_metrics,
            self._symexec_font_height,
            self._symexec_font_width,
            self._symexec_font_ascent,
        ) = ConfigurationManager._manage_font_cache(
            self.symexec_font,
            self._symexec_font,
            self._symexec_font_metrics,
            self._symexec_font_height,
            self._symexec_font_width,
            self._symexec_font_ascent,
        )

    def _code_manage_font_cache(self) -> None:
        (
            self._code_font,
            self._code_font_metrics,
            self._code_font_height,
            self._code_font_width,
            self._code_font_ascent,
        ) = ConfigurationManager._manage_font_cache(
            self.code_font,
            self._code_font,
            self._code_font_metrics,
            self._code_font_height,
            self._code_font_width,
            self._code_font_ascent,
        )

    disasm_font: QFont
    symexec_font: QFont
    code_font: QFont

    @property
    def disasm_font_metrics(self):
        self._disasm_manage_font_cache()
        return self._disasm_font_metrics

    @property
    def disasm_font_height(self):
        self._disasm_manage_font_cache()
        return self._disasm_font_height

    @property
    def disasm_font_width(self):
        self._disasm_manage_font_cache()
        return self._disasm_font_width

    @property
    def disasm_font_ascent(self):
        self._disasm_manage_font_cache()
        return self._disasm_font_ascent

    @property
    def symexec_font_metrics(self):
        self._symexec_manage_font_cache()
        return self._symexec_font_metrics

    @property
    def symexec_font_height(self):
        self._symexec_manage_font_cache()
        return self._symexec_font_height

    @property
    def symexec_font_width(self):
        self._symexec_manage_font_cache()
        return self._symexec_font_width

    @property
    def symexec_font_ascent(self):
        self._symexec_manage_font_cache()
        return self._symexec_font_ascent

    @property
    def code_font_metrics(self):
        self._code_manage_font_cache()
        return self._code_font_metrics

    @property
    def code_font_height(self):
        self._code_manage_font_cache()
        return self._code_font_height

    @property
    def code_font_width(self):
        self._code_manage_font_cache()
        return self._code_font_width

    @property
    def code_font_ascent(self):
        self._code_manage_font_cache()
        return self._code_font_ascent

    def init_font_config(self) -> None:
        if self.ui_default_font is None:
            self.ui_default_font = QApplication.font("QMenu")
        if self.tabular_view_font is None:
            self.tabular_view_font = QApplication.font("QMenu")

    recent_files: list[str]

    def recent_file(self, file_path: str) -> None:
        with contextlib.suppress(ValueError):
            self.recent_files.remove(file_path)
        self.recent_files = self.recent_files[:9]
        self.recent_files.append(file_path)

    def __getattr__(self, item):
        if item in self.__slots__ or item in type(self).__dict__:
            return super().__getattribute__(item)

        if item in self._entries:
            return self._entries[item].value

        raise AttributeError(item)

    def __setattr__(self, key, value) -> None:
        if key in self.__slots__ or key in type(self).__dict__:
            super().__setattr__(key, value)
            return

        if key in self._entries:
            self._entries[key].value = value
            return

        raise AttributeError(key)

    def connect(self, key: str, func: Callable[[Any], None], init: bool) -> None:
        """
        Connect func to the QT signal emitted when the key changes
        If init, calls func on the value associated with key after connecting
        """
        self._entries[key].changed.connect(func)
        if init:
            func(getattr(self, key))

    def disconnect(self, key: str, func: Callable[[Any], None]) -> None:
        """
        Disconnect func from the QT signal emitted when the key changes
        """
        self._entries[key].changed.disconnect(func)

    def __dir__(self):
        return list(super().__dir__()) + list(self._entries)

    @classmethod
    def parse(cls, f, ignore_unknown_entries: bool = False):
        entry_map = {}
        for entry in ENTRIES:
            entry_map[entry.name] = entry.copy()

        try:
            loaded = tomlkit.load(f)

            for k, v in loaded.items():
                if k not in entry_map:
                    if ignore_unknown_entries:
                        _l.warning("Unknown configuration option '%s'. Ignoring...", k)
                    else:
                        entry_map[k] = UninterpretedCE(k, v)
                    continue
                    # default to a string

                entry = entry_map[k]
                entry.value = cls.deserialize(entry.type_, k, v)
                if entry.value is None:
                    entry_map[k] = UninterpretedCE(k, v)
        except tomlkit.exceptions.ParseError:
            _l.error("Failed to parse configuration file: '%s'. Continuing with default options...", exc_info=True)

        return cls(entry_map)

    @staticmethod
    def deserialize(ty, k, v):
        if ty in data_serializers:
            v = data_serializers[ty][0](k, v)
            if v is None:
                return None
        else:
            try:
                v = tomltype2pytype(v, ty)
            except TypeError:
                _l.warning(
                    "Value '%s' for configuration option '%s' has type '%s', expected type '%s'. Ignoring...",
                    v,
                    k,
                    type(v),
                    ty,
                )
                return None

        return v

    def reinterpet(self) -> None:
        """
        Walks the ENTRIES list, trying to update self's entries with respect to anything that may have been added to
        the global list. Tries to fix up UninterpretedCEs. Should be called e.g. after loading plugins.
        """
        for entry in ENTRIES:
            my_entry = self._entries.get(entry.name, None)
            if my_entry is None:
                self._entries[entry.name] = my_entry
                continue

            if type(my_entry) is UninterpretedCE:
                entry.value = self.deserialize(entry.type_, entry.name, my_entry.value)
                if entry.value is not None:
                    self._entries[entry.name] = entry

    @classmethod
    def parse_file(cls, path: str, ignore_unknown_entries: bool = False):
        with open(path, encoding="utf-8") as f:
            return cls.parse(f, ignore_unknown_entries=ignore_unknown_entries)

    def save(self, f) -> None:
        out = {}
        for k, v in self._entries.items():
            v = v.value
            while type(v) in data_serializers:
                v = data_serializers[type(v)][1](k, v)
            out[k] = v

        tomlkit.dump(out, f)

    def save_file(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as f:
            self.save(f)

    def attempt_importing_initial_config(self) -> bool:
        """
        Look for am_initial_config inside the last four levels of directories. Import the first one found. Then remove
        the file. Prompt user to manually remove the file if file removing fails.

        :return: True if successfully imports the initial configuration. False otherwise.
        """

        loaded = False

        base = app_root()
        for _ in range(4):
            initial_config_path = os.path.join(base, "am_initial_config")
            if os.path.isfile(initial_config_path):
                from . import save_config  # delayed import # pylint: disable=import-outside-toplevel

                # we found it!
                new_conf = self.__class__.parse_file(initial_config_path, ignore_unknown_entries=False)
                # copy entries over
                self._entries = new_conf._entries
                # save it!
                save_config()
                loaded = True

                # remove the file
                try:
                    os.remove(initial_config_path)
                except (IsADirectoryError, FileNotFoundError):
                    pass
                except Exception:  # pylint: disable=broad-except
                    QMessageBox.warning(
                        None,
                        "Failed to remove the initial configuration file",
                        f"angr management imported the initial configuration but failed to remove the"
                        f"initial configuration file at {initial_config_path}. Please remove it "
                        f"manually. Otherwise your settings will be overwritten next time angr "
                        f"management starts.",
                    )

                break

            last_dirname = base
            base = os.path.dirname(last_dirname)

            if base == last_dirname:
                # we reached the end of the directory hierarchy
                break

        return loaded

    @property
    def has_operation_mango(self) -> bool:
        try:
            import argument_resolver  # noqa

            return True
        except ImportError:
            return False



from __future__ import annotations

from .analysis_options import AnalysisOptionsDialog
from .assemble_patch import AssemblePatchDialog
from .breakpoint import BreakpointDialog
from .load_binary import LoadBinary
from .load_docker_prompt import LoadDockerPrompt
from .load_plugins import LoadPlugins
from .preferences import Preferences

__all__ = [
    "AnalysisOptionsDialog",
    "AssemblePatchDialog",
    "BreakpointDialog",
    "LoadBinary",
    "LoadDockerPrompt",
    "LoadPlugins",
    "Preferences",
]



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from angrmanagement.config import Conf

from .menu import Menu, MenuEntry, MenuSeparator

if TYPE_CHECKING:
    from collections.abc import Callable


class DisasmLabelContextMenu(Menu):
    def __init__(self, disasm_view) -> None:
        super().__init__("", parent=disasm_view)

        self.addr: int = None

        self.entries.extend(
            [
                MenuEntry("T&oggle selection", self._toggle_label_selection),
                MenuSeparator(),
                MenuEntry("&XRefs...", self._popup_xrefs),
                MenuSeparator(),
            ]
        )
        if Conf.has_operation_mango:
            self.entries.extend(
                [
                    MenuEntry("&Depends on...", self._popup_dependson_dialog),
                    MenuSeparator(),
                ]
            )
        self.entries.extend(
            [
                MenuEntry("E&xecute symbolically...", self._popup_newstate_dialog),
                MenuEntry("&Avoid in execution...", self._avoid_in_execution),
                MenuEntry("&Find in execution...", self._find_in_execution),
            ]
        )

    @property
    def _disasm_view(self):
        return self.parent

    def _popup_newstate_dialog(self) -> None:
        self._disasm_view.popup_newstate_dialog(async_=True)

    def _popup_dependson_dialog(self) -> None:
        self._disasm_view.popup_dependson_dialog(addr=self.addr, func=True)

    def _toggle_label_selection(self) -> None:
        self._disasm_view.infodock.toggle_label_selection(self.addr)

    def _avoid_in_execution(self) -> None:
        self._disasm_view.avoid_addr_in_exec(self.addr)

    def _find_in_execution(self) -> None:
        self._disasm_view.find_addr_in_exec(self.addr)

    def _popup_xrefs(self) -> None:
        if self._disasm_view is None or self._disasm_view._flow_graph is None:
            return
        self._disasm_view.popup_xref_dialog(addr=self.addr, variable=None, dst_addr=self.addr, async_=True)

    #
    # Public Methods
    #

    def add_menu_entry(
        self, text: str, callback: Callable[[DisasmLabelContextMenu], None], add_separator_first: bool = True
    ) -> None:
        if add_separator_first:
            self.entries.append(MenuSeparator())
        self.entries.append(MenuEntry(text, partial(callback, self)))

    def remove_menu_entry(self, text: str, remove_preceding_separator: bool = True) -> None:
        for idx, m in enumerate(self.entries):
            if not isinstance(m, MenuEntry):
                continue
            if m.caption == text:
                self.entries.remove(m)
                if remove_preceding_separator:
                    self.entries.pop(idx - 1)



from __future__ import annotations

from typing import TYPE_CHECKING

import claripy
from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class QASTViewer(QFrame):
    def __init__(
        self,
        ast,
        workspace: Workspace | None = None,
        custom_painting: bool = False,
        display_size: bool = True,
        byte_format=None,
        parent=None,
    ) -> None:
        super().__init__(parent)

        # configs
        self._custom_painting = custom_painting
        self._ast = ast
        self._display_size = display_size
        self._byte_format = byte_format

        # string representations for display
        self._size_str = None
        self._ast_str = None

        # properties that are only used in custom painting mode
        self._x = None
        self._y = None
        self._width = None
        self._height = None

        # widgets. only used in normal painting mode
        self._size_label = None
        self._ast_label = None

        self.setFrameShape(QFrame.NoFrame)
        self.setLineWidth(0)

        # workspace backref
        self.workspace = workspace

        if not self._custom_painting:
            self._init_widgets()
        else:
            self.reload()

    def mouseDoubleClickEvent(self, event) -> None:
        if self._ast is not None and not self._ast.symbolic:
            self.workspace.viz(self._ast._model_concrete.value)

    #
    # Properties
    #

    @property
    def ast(self):
        return self._ast

    @ast.setter
    def ast(self, v) -> None:
        self._ast = v
        self.reload()

    @property
    def x(self):
        if not self._custom_painting:
            raise ValueError("QASTViewer does not have a size when custom painting is disabled.")
        return self._x

    @x.setter
    def x(self, v):
        if not self._custom_painting:
            raise ValueError("QASTViewer does not have a size when custom painting is disabled.")
        self._x = v

    @property
    def y(self):
        if not self._custom_painting:
            raise ValueError("QASTViewer does not have a size when custom painting is disabled.")
        return self._y

    @y.setter
    def y(self, v):
        if not self._custom_painting:
            raise ValueError("QASTViewer does not have a size when custom painting is disabled.")
        self._y = v

    @property
    def width(self):
        if not self._custom_painting:
            return super().width()
        else:
            return self._width

    @property
    def height(self):
        if not self._custom_painting:
            return super().height()
        else:
            return self._height

    #
    # Public methods
    #

    def paint(self, painter) -> None:
        """

        :param QPainter painter:
        :return:
        """

        if self.x is None or self.y is None:
            # paint() is called before x and y are set
            return

        painter.drawText(self.x, self.y + Conf.symexec_font_ascent, self._ast_str)

    def reload(self) -> None:
        # build string representations
        self._build_strings()

        if not self._custom_painting:
            self._reload_widgets()
        else:
            self._determine_size()

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        layout = QHBoxLayout()

        ast_label = QLabel(self)
        self._ast_label = ast_label

        if self._display_size:
            size_label = QLabel(self)
            size_label.setProperty("class", "ast_viewer_size")
            size_label.setAlignment(Qt.AlignRight)
            size_label.setMaximumSize(QSize(24, 65536))
            self._size_label = size_label
            layout.addWidget(self._size_label)

        if self._ast is not None:
            self.reload()

        layout.addWidget(ast_label)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def _reload_widgets(self) -> None:
        if self._ast is None:
            self._ast_label.setText("")
            return

        ast = self._ast

        # set style
        if isinstance(ast, int) or not ast.symbolic:
            self._ast_label.setProperty("class", "ast_viewer_ast_concrete")
        else:
            self._ast_label.setProperty("class", "ast_viewer_ast_symbolic")

        # set text
        if self._display_size:
            self._size_label.setText(self._size_str)
        else:
            self._size_label.setText("")

        self._ast_label.setText(self._ast_str)

        # reapply the style
        self._ast_label.style().unpolish(self._ast_label)
        self._ast_label.style().polish(self._ast_label)

    def _build_strings(self) -> None:
        if self._ast is None:
            self._ast_label.setText("")
            return

        ast = self._ast

        # set text
        if isinstance(ast, int):
            if self._display_size:
                self._size_str = "[Unknown]"
            format = "%02x" if self._byte_format is None else self._byte_format
            self._ast_str = format % ast
        else:
            # claripy.AST
            if self._display_size:
                self._size_label.setText("[%d]" % (len(ast) // 8))  # in bytes
            if not ast.symbolic:
                format = "%02x" if self._byte_format is None else self._byte_format
                self._ast_str = format % self._ast._model_concrete.value
            else:
                # symbolic
                if isinstance(ast, claripy.ast.BV) and ast.op == "BVS":
                    var_name = ast.args[0]
                    self._ast_str = var_name
                else:
                    self._ast_str = ast.__repr__(max_depth=1)

    def _determine_size(self) -> None:
        self._height = Conf.symexec_font_height
        self._width = Conf.symexec_font_width * len(self._ast_str)
        if self._display_size:
            self._width += Conf.symexec_font_width * len(self._size_str)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout

from angrmanagement.ui.widgets.qstate_table import QStateTable

from .view import InstanceView

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class StatesView(InstanceView):
    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("states", workspace, default_docking_position, instance)

        self.base_caption = "States"
        self._state_table: QStateTable

        self._init_widgets()

    def reload(self) -> None:
        self._state_table.state_manager = self.instance.states

    def closeEvent(self, event) -> None:
        """
        Close children before exiting
        """
        self._state_table.close()

    def sizeHint(self):
        return QSize(400, 800)

    def _init_widgets(self) -> None:
        self._state_table = QStateTable(self.workspace, self.instance, self, selection_callback=self._on_state_selected)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self._state_table)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

    def _on_state_selected(self, state) -> None:
        """
        A new function is on selection right now. Update the disassembly view that is currently at front.

        :param function:
        :return:
        """



from __future__ import annotations

from .menu import Menu, MenuEntry


class DisasmOptionsMenu(Menu):
    def __init__(self, disasm_view) -> None:
        super().__init__("", parent=disasm_view)

        self._show_minimap_action = MenuEntry(
            "Show &minimap", self._show_minimap, checkable=True, checked=self.parent.show_minimap
        )
        self._smart_highlighting_action = MenuEntry(
            "Smart &highlighting", self._smart_highlighting, checkable=True, checked=self.parent.smart_highlighting
        )
        self._show_address_action = MenuEntry(
            "Show &address", self._show_address, checkable=True, checked=self.parent.show_address
        )
        self._show_variable_action = MenuEntry(
            "Show &variable", self._show_variable, checkable=True, checked=self.parent.show_variable
        )
        self._show_variable_ident_action = MenuEntry(
            "Show variable &identifiers",
            self._show_variable_identifier,
            checkable=True,
            checked=self.parent.show_variable_identifier,
        )
        self._show_exception_edges_action = MenuEntry(
            "Show &exception transition edges",
            self._show_exception_edges,
            checkable=True,
            checked=self.parent.show_exception_edges,
        )

        self.entries.extend(
            [
                self._show_minimap_action,
                self._smart_highlighting_action,
                self._show_address_action,
                self._show_variable_action,
                self._show_variable_ident_action,
                self._show_exception_edges_action,
            ]
        )

    def _show_minimap(self) -> None:
        checked = self._show_minimap_action.checked
        self.parent.toggle_show_minimap(checked)

    def _smart_highlighting(self) -> None:
        checked = self._smart_highlighting_action.checked
        self.parent.toggle_smart_highlighting(checked)

    def _show_address(self) -> None:
        checked = self._show_address_action.checked
        self.parent.toggle_show_address(checked)

    def _show_variable(self) -> None:
        checked = self._show_variable_action.checked
        self.parent.toggle_show_variable(checked)

    def _show_variable_identifier(self) -> None:
        checked = self._show_variable_ident_action.checked
        self.parent.toggle_show_variable_identifier(checked)

    def _show_exception_edges(self) -> None:
        checked = self._show_exception_edges_action.checked
        self.parent.toggle_show_exception_edges(checked)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QVBoxLayout

from angrmanagement.ui.views.view import InstanceView
from angrmanagement.ui.widgets import QTraceMap

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class TraceMapView(InstanceView):
    """
    View container for QTraceMap.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("tracemap", workspace, default_docking_position, instance)
        self.base_caption: str = "Trace Map"
        self.inner_widget: QTraceMap | None = None
        self._init_widgets()

    @staticmethod
    def minimumSizeHint():
        return QSize(25, 25)

    @staticmethod
    def sizeHint():
        return QSize(25, 25)

    def _init_widgets(self) -> None:
        """
        Initialize widgets for this view.
        """
        self.inner_widget = QTraceMap(self.instance, parent=self)
        lyt = QVBoxLayout()
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(self.inner_widget)
        self.setLayout(lyt)
